"""
PDF content objects created by the interpreter.
"""

import itertools
import logging
import operator
from abc import abstractmethod
from copy import copy
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Sequence,
    Sized,
    Tuple,
    Type,
    Union,
    overload,
)

from playa.color import (
    BASIC_BLACK,
    LITERAL_DEFAULT,
    LITERAL_NORMAL,
    LITERAL_RELATIVE_COLORIMETRIC,
    PREDEFINED_COLORSPACE,
    Color,
    ColorSpace,
)
from playa.font import CIDFont, Font, Type3Font
from playa.parser import LIT, ContentParser, Lexer, Token
from playa.pdftypes import (
    ContentStream,
    Matrix,
    PDFObject,
    Point,
    PSKeyword,
    PSLiteral,
    Rect,
    dict_value,
    int_value,
    list_value,
    matrix_value,
    rect_value,
    resolve1,
)
from playa.utils import (
    apply_matrix_pt,
    decode_text,
    get_bound,
    get_bound_rects,
    mult_matrix,
    transform_bbox,
    translate_matrix,
    update_glyph_offset,
)
from playa.worker import PageRef, _deref_document, _deref_page

if TYPE_CHECKING:
    from playa.document import Document
    from playa.page import Page
    from playa.structure import Element, PageStructure

log = logging.getLogger(__name__)


class DashPattern(NamedTuple):
    """
    Line dash pattern in PDF graphics state (PDF 1.7 section 8.4.3.6).

    Attributes:
      dash: lengths of dashes and gaps in user space units
      phase: starting position in the dash pattern
    """

    dash: Tuple[float, ...]
    phase: float

    def __str__(self):
        if len(self.dash) == 0:
            return ""
        else:
            return f"{self.dash} {self.phase}"


SOLID_LINE = DashPattern((), 0)


@dataclass
class GraphicState:
    """PDF graphics state (PDF 1.7 section 8.4) including text state
    (PDF 1.7 section 9.3.1), but excluding coordinate transformations.

    Contrary to the pretensions of pdfminer.six, the text state is for
    the most part not at all separate from the graphics state, and can
    be updated outside the confines of `BT` and `ET` operators, thus
    there is no advantage and only confusion that comes from treating
    it separately.

    The only state that does not persist outside `BT` / `ET` pairs is
    the text coordinate space (line matrix and text rendering matrix),
    and it is also the only part that is updated during iteration over
    a `TextObject`.

    For historical reasons the main coordinate transformation matrix,
    though it is also part of the graphics state, is also stored
    separately.

    Attributes:
      clipping_path: The current clipping path (sec. 8.5.4)
      linewidth: Line width in user space units (sec. 8.4.3.2)
      linecap: Line cap style (sec. 8.4.3.3)
      linejoin: Line join style (sec. 8.4.3.4)
      miterlimit: Maximum length of mitered line joins (sec. 8.4.3.5)
      dash: Dash pattern for stroking (sec 8.4.3.6)
      intent: Rendering intent (sec. 8.6.5.8)
      stroke_adjustment: A flag specifying whether to compensate for
        possible rasterization effects when stroking a path with a line
        width that is small relative to the pixel resolution of the output
        device (sec. 10.7.5)
      blend_mode: The current blend mode that shall be used in the
        transparent imaging model (sec. 11.3.5)
      smask: A soft-mask dictionary (sec. 11.6.5.1) or None
      salpha: The constant shape or constant opacity value used for
        stroking operations (sec. 11.3.7.2 & 11.6.4.4)
      nalpha: The constant shape or constant opacity value used for
        non-stroking operations
      alpha_source: A flag specifying whether the current soft mask and
        alpha constant parameters shall be interpreted as shape values
        (true) or opacity values (false). This flag also governs the
        interpretation of the SMask entry, if any, in an image dictionary
      black_pt_comp: The black point compensation algorithm that shall be
        used when converting CIE-based colours (sec. 8.6.5.9)
      flatness: The precision with which curves shall be rendered on
        the output device (sec. 10.6.2)
      scolor: Colour used for stroking operations
      scs: Colour space used for stroking operations
      ncolor: Colour used for non-stroking operations
      ncs: Colour space used for non-stroking operations
      font: The current font.
      fontsize: The "font size" parameter, which is **not** the font
        size in points as you might understand it, but rather a
        scaling factor applied to text space (so, it affects not only
        text size but position as well).  Since most reasonable people
        find that behaviour rather confusing, this is often just 1.0,
        and PDFs rely on the text matrix to set the size of text.
      charspace: Extra spacing to add after each glyph, expressed in
        unscaled text space units, meaning it is not affected by
        `fontsize`.  **BUT** it will be modified by `scaling` for
        horizontal writing mode (so, most of the time).
      wordspace: Extra spacing to add after a space glyph, defined
        very specifically as the glyph encoded by the single-byte
        character code 32 (SPOILER: it is probably a space).  Also
        expressed in unscaled text space units, but modified by
        `scaling`.
      scaling: The horizontal scaling factor as defined by the PDF
        standard (that is, divided by 100).
      leading: The leading as defined by the PDF standard, in unscaled
        text space units.
      render_mode: The PDF rendering mode.  The really important one
        here is 3, which means "don't render the text".  You might
        want to use this to detect invisible text.
      rise: The text rise (superscript or subscript position), in
        unscaled text space units.
      knockout: The text knockout flag, shall determine the behaviour of
        overlapping glyphs within a text object in the transparent imaging
        model (sec. 9.3.8)

    """

    clipping_path: None = None  # TODO
    linewidth: float = 1
    linecap: int = 0
    linejoin: int = 0
    miterlimit: float = 10
    dash: DashPattern = SOLID_LINE
    intent: PSLiteral = LITERAL_RELATIVE_COLORIMETRIC
    stroke_adjustment: bool = False
    blend_mode: Union[PSLiteral, List[PSLiteral]] = LITERAL_NORMAL
    smask: Union[None, Dict[str, PDFObject]] = None
    salpha: float = 1
    nalpha: float = 1
    alpha_source: bool = False
    black_pt_comp: PSLiteral = LITERAL_DEFAULT
    flatness: float = 1
    scolor: Color = BASIC_BLACK
    scs: ColorSpace = PREDEFINED_COLORSPACE["DeviceGray"]
    ncolor: Color = BASIC_BLACK
    ncs: ColorSpace = PREDEFINED_COLORSPACE["DeviceGray"]
    font: Union[Font, None] = None
    fontsize: float = 0
    charspace: float = 0
    wordspace: float = 0
    scaling: float = 100
    leading: float = 0
    render_mode: int = 0
    rise: float = 0
    knockout: bool = True


class MarkedContent(NamedTuple):
    """Marked content point or section in a PDF page or Form XObject.

    Attributes:
      mcid: Marked content section ID, or `None` for a marked content point.
      tag: Name of tag for this marked content.
      props: Marked content property dictionary.

    """

    mcid: Union[int, None]
    tag: str
    props: Dict[str, PDFObject]


PathOperator = Literal["h", "m", "l", "v", "c", "y"]


class PathSegment(NamedTuple):
    """
    Segment in a PDF graphics path.
    """

    operator: PathOperator
    points: Tuple[Point, ...]


@dataclass
class ContentObject(Sized, Iterable["ContentObject"]):
    """Any sort of content object.

    Attributes:
      gstate: Graphics state.
      ctm: Coordinate transformation matrix (PDF 1.7 section 8.3.2).
      mcstack: Stack of enclosing marked content sections.
    """

    _pageref: PageRef
    _parentkey: Union[int, None]
    gstate: GraphicState
    ctm: Matrix
    mcstack: Tuple[MarkedContent, ...]

    def __iter__(self) -> Iterator["ContentObject"]:
        yield from ()

    def __len__(self) -> int:
        """Return the number of children of this object (generic implementation)."""
        return sum(1 for _ in self)

    def finalize(self) -> "ContentObject":
        """Finalize this content object so it can be reused.

        Because certain elements (specifically the graphics state) are
        references to mutable objects that are updated during content
        parsing, reusing ContentObjects is fraught with peril.  This
        was a conscious design choice to save time and memory, but
        perhaps not a good one.

        If you want to reuse a ContentObject, thus, you must call this
        method on it.  For convenience, it returns the same
        ContentObject, allowing you to do, for instance:

            objs = [x.finalize() for x in page]
        """
        self.gstate = copy(self.gstate)
        return self

    @property
    def object_type(self) -> str:
        """Type of this object as a string, e.g. "text", "path", "image"."""
        name = self.__class__.__name__
        return name[: -len("Object")].lower()

    @property
    def bbox(self) -> Union[Rect, None]:
        """The bounding box in device space of this object."""
        # These bboxes have already been computed in device space so
        # we don't need all 4 corners!
        return get_bound_rects(item.bbox for item in self)

    @property
    def mcs(self) -> Union[MarkedContent, None]:
        """The immediately enclosing marked content section."""
        return self.mcstack[-1] if self.mcstack else None

    @property
    def mcid(self) -> Union[int, None]:
        """The marked content ID of the nearest enclosing marked
        content section with an ID.

        This is notably what you should use (and what `parent` uses)
        to find the parent logical structure element, because (PDF
        14.7.5.1.1):

        > A marked-content sequence corresponding to a structure
        content item shall not have another marked-content sequence
        for a structure content item nested within it though
        non-structural marked-content shall be allowed.
        """
        if hasattr(self, "_mcid"):
            return self._mcid
        for mcs in self.mcstack[::-1]:
            if mcs.mcid is not None:
                self._mcid: Union[int, None] = mcs.mcid
                break
        else:
            self._mcid = None
        return self._mcid

    @property
    def parent(self) -> Union["Element", None]:
        """The enclosing logical structure element, if any."""
        from playa.structure import Element

        # Use `mcid` and not `mcs` here (see docs for `mcid`)
        if hasattr(self, "_parent"):
            return self._parent
        self._parent: Union["Element", None] = None
        parent_key = self._parentkey
        if parent_key is None:
            return self._parent
        structure = self.doc.structure
        if structure is None:
            return self._parent
        mcid = self.mcid
        if mcid is None:
            return self._parent
        parents = list_value(structure.parent_tree[parent_key])
        if mcid >= len(parents):
            log.warning(
                "Invalid marked content ID: %d (page has %d MCIDs)", mcid, len(parents)
            )
            return self._parent
        if parents[mcid] is None:  # This might mean we have the wrong StructParents
            log.warning(
                "Marked content ID %d on page_idx %d has no parent element",
                mcid,
                self.page.page_idx,
            )
            return self._parent
        self._parent = Element.from_dict(self.doc, dict_value(parents[mcid]))
        return self._parent

    @property
    def page(self) -> "Page":
        """The page containing this content object."""
        return _deref_page(self._pageref)

    @property
    def doc(self) -> "Document":
        """The document containing this content object."""
        docref, _ = self._pageref
        return _deref_document(docref)


@dataclass
class TagObject(ContentObject):
    """A marked content tag.."""

    _mcs: MarkedContent

    def __len__(self) -> int:
        """A tag has no contents, iterating over it returns nothing."""
        return 0

    @property
    def mcs(self) -> MarkedContent:
        """The marked content tag for this object."""
        return self._mcs

    @property
    def mcid(self) -> Union[int, None]:
        """The marked content ID of the nearest enclosing marked
        content section with an ID."""
        if self._mcs.mcid is not None:
            return self._mcs.mcid
        return super().mcid

    @property
    def bbox(self) -> Union[Rect, None]:
        """A tag has no content and thus no bounding box."""
        return None


@dataclass
class ImageObject(ContentObject):
    """An image (either inline or XObject).

    Attributes:
      xobjid: Name of XObject (or None for inline images).
      srcsize: Size of source image in pixels.
      bits: Number of bits per component, if required (otherwise 1).
      imagemask: True if the image is a mask.
      stream: Content stream with image data.
      colorspace: Colour space for this image, if required (otherwise
        None).
    """

    xobjid: Union[str, None]
    srcsize: Tuple[int, int]
    bits: int
    imagemask: bool
    stream: ContentStream
    colorspace: Union[ColorSpace, None]

    def __contains__(self, name: str) -> bool:
        return name in self.stream

    def __getitem__(self, name: str) -> PDFObject:
        return self.stream[name]

    def get(self, name: str, default: PDFObject = None) -> PDFObject:
        return self.stream.get(name, default)

    def __len__(self) -> int:
        """Even though you can __getitem__ from an image you cannot iterate
        over its keys, sorry about that.  Returns zero."""
        return 0

    @property
    def parent(self) -> Union["Element", None]:
        """The enclosing logical structure element, if any."""
        from playa.structure import Element

        if hasattr(self, "_parent"):
            return self._parent
        self._parent = None
        if self._parentkey is None:
            return self._parent
        # No structure, no parent!
        if self.doc.structure is None:
            return self._parent
        try:
            parent = resolve1(self.doc.structure.parent_tree[self._parentkey])
            if isinstance(parent, dict):
                self._parent = Element.from_dict(self.doc, parent)
            else:
                del self._parent
                return super().parent
        except IndexError:
            pass
        return self._parent

    @property
    def buffer(self) -> bytes:
        """Binary stream content for this image"""
        return self.stream.buffer

    @property
    def bbox(self) -> Union[Rect, None]:
        # PDF 1.7 sec 8.3.24: All images shall be 1 unit wide by 1
        # unit high in user space, regardless of the number of samples
        # in the image. To be painted, an image shall be mapped to a
        # region of the page by temporarily altering the CTM.
        return transform_bbox(self.ctm, (0, 0, 1, 1))


# Group XObject subtypes. As of PDF 2.0 Transparency is the only defined subtype
LITERAL_TRANSPARENCY = LIT("Transparency")


@dataclass
class XObjectObject(ContentObject):
    """An eXternal Object, in the context of a page.

    There are a couple of kinds of XObjects.  Here we are only
    concerned with "Form XObjects" which, despite their name, have
    nothing at all to do with fillable forms.  Instead they are like
    little embeddable PDF pages, possibly with their own resources,
    definitely with their own definition of "user space".

    Image XObjects are handled by `ImageObject`.

    Attributes:
      xobjid: Name of this XObject (in the page resources).
      stream: Content stream with PDF operators.
      resources: Resources specific to this XObject, if any.
      group: Transparency group, if any.
    """

    xobjid: str
    stream: ContentStream
    resources: Union[None, Dict[str, PDFObject]]
    group: Union[None, Dict[str, PDFObject]]

    @property
    def bbox(self) -> Rect:
        """Get the bounding box of this XObject in device space."""
        # It is a required attribute!
        if "BBox" not in self.stream:
            log.debug("XObject %r has no BBox: %r", self.xobjid, self.stream)
            return self.page.cropbox
        return transform_bbox(self.ctm, rect_value(self.stream["BBox"]))

    @property
    def buffer(self) -> bytes:
        """Raw stream content for this XObject"""
        return self.stream.buffer

    @property
    def tokens(self) -> Iterator[Token]:
        """Iterate over tokens in the XObject's content stream."""
        for _, tok in Lexer(self.stream.buffer):
            yield tok

    @property
    def contents(self) -> Iterator[PDFObject]:
        """Iterator over PDF objects in the content stream."""
        for pos, obj in ContentParser([self.stream], self.doc):
            yield obj

    def interp(
        self,
        filter_classes: Union[Collection[Type[ContentObject]], None] = None,
        restrict_ops: Union[Collection[PSKeyword], None] = None,
    ) -> Iterator[ContentObject]:
        from playa.interp import LazyInterpreter

        return LazyInterpreter(
            self.page,
            [self.stream],
            self.resources,
            ctm=self.ctm,
            gstate=self.gstate,
            # This is not really correct if this XObject has a
            # StructParent, *but* in that case the standard forbids
            # there to be any marked content sections inside it so
            # this should never get accessed anyway.
            parent_key=self._parentkey,
            filter_classes=filter_classes,
            restrict_ops=restrict_ops,
        )

    def __iter__(self) -> Iterator[ContentObject]:
        return self.interp()

    @property
    def parent(self) -> Union["Element", None]:
        """The enclosing logical structure element, if any."""
        from playa.structure import Element

        if hasattr(self, "_parent"):
            return self._parent
        self._parent = None
        if self._parentkey is None:
            return self._parent
        # No structure, no parent!
        if self.doc.structure is None:
            return self._parent
        try:
            parent = resolve1(self.doc.structure.parent_tree[self._parentkey])
            if isinstance(parent, dict):
                self._parent = Element.from_dict(self.doc, parent)
            else:
                del self._parent
                return super().parent
        except IndexError:
            pass
        return self._parent

    @property
    def structure(self) -> "PageStructure":
        """Mapping of marked content IDs to logical structure elements.

        As with pages, Form XObjects can also contain their own
        mapping of marked content IDs to structure elements.
        """
        from playa.structure import PageStructure

        if hasattr(self, "_structmap"):
            return self._structmap
        self._structmap: PageStructure = PageStructure(self._pageref, [])
        if self.doc.structure is None:
            return self._structmap
        if self._parentkey is None:
            return self._structmap
        try:
            self._structmap = PageStructure(
                self._pageref, self.doc.structure.parent_tree[self._parentkey]
            )
        except IndexError as e:
            log.warning("Invalid StructParents: %r (%s)", self._parentkey, e)
        except TypeError:
            # This means that there is a single StructParent, and thus
            # no internal structure to this Form XObject
            pass
        return self._structmap

    @property
    def marked_content(self) -> "ContentSequence":
        """A [`ContentSequence`][playa.content.ContentSequence] containing
        content objects associated with the structural elements in
        [`structure`][playa.content.XObjectObject.structure].  They
        consist of a sequence with the same indices (these are the
        marked content IDs) as the structure so can be zipped:

            for element, contents in zip(xobj.structure,
                                         xobj.marked_content):
                for obj in contents:
                    ...  # do something with it

        Or you can also access the contents of a single element:

            for obj in xobj.marked_content[mcid]:
                ... # do something with it

        """
        if hasattr(self, "_marked_contents"):
            return self._marked_contents
        self._marked_contents: ContentSequence = ContentSequence(self)
        return self._marked_contents

    @property
    def fonts(self) -> Mapping[str, Font]:
        """Mapping of resource names to fonts for this Form XObject.

        Note: This is not the same as `playa.Document.fonts`.
            The resource names (e.g. `F1`, `F42`, `FooBar`) here are
            specific to a page (or Form XObject) resource dictionary
            and have no relation to the font name as commonly
            understood (e.g. `Helvetica`,
            `WQERQE+Arial-SuperBold-HJRE-UTF-8`).  Since font names are
            generally considered to be globally unique, it may be
            possible to access fonts by them in the future.

        """
        from playa.interp import FontMapping

        if hasattr(self, "_fontmap"):
            return self._fontmap
        self._fontmap: Mapping[str, Font] = FontMapping(
            self.resources.get("Font") if self.resources else None, self.doc
        )
        return self._fontmap

    @classmethod
    def from_stream(
        cls,
        stream: ContentStream,
        page: "Page",
        xobjid: str,
        gstate: GraphicState,
        ctm: Matrix,
        mcstack: Tuple[MarkedContent, ...],
    ) -> "XObjectObject":
        """Create a new XObjectObject from a content stream."""
        if "Matrix" in stream:
            ctm = mult_matrix(matrix_value(stream["Matrix"]), ctm)
        # According to PDF reference 1.7 section 4.9.1, XObjects in
        # earlier PDFs (prior to v1.2) use the page's Resources entry
        # instead of having their own Resources entry.  So, this could
        # be None, in which case LazyInterpreter will fall back to
        # page.resources.
        xobjres = stream.get("Resources")
        resources = None if xobjres is None else dict_value(xobjres)
        xobjgrp = stream.get("Group")
        group = None if xobjgrp is None else dict_value(xobjgrp)
        # PDF 2.0, sec 11.6.6
        # Initial blend mode: Before execution of the transparency group
        # XObject’s content stream, the current blend mode in the graphics
        # state shall be initialised to Normal, the current stroking and
        # nonstroking alpha constants to 1.0, and the current soft mask to None
        if group and group.get("S") == LITERAL_TRANSPARENCY:
            # Need to copy here so as not to modify existing gstate,
            # unfortunately it will get copied again later...
            gstate = copy(gstate)
            gstate.blend_mode = LITERAL_NORMAL
            gstate.salpha = gstate.nalpha = 1
            gstate.smask = None
        # PDF 2.0, Table 359
        # At most one of [StructParent and StructParents] shall be
        # present in a given object. An object may be either a content
        # item in its entirety or a container for marked-content
        # sequences that are content items, but not both.
        if "StructParent" in stream:
            parent_key = int_value(stream["StructParent"])
        elif "StructParents" in stream:
            parent_key = int_value(stream["StructParents"])
        else:
            parent_key = None
        return cls(
            _pageref=page.pageref,
            _parentkey=parent_key,
            gstate=gstate,
            ctm=ctm,
            mcstack=mcstack,
            xobjid=xobjid,
            stream=stream,
            resources=resources,
            group=group,
        )


@dataclass
class PathObject(ContentObject):
    """A path object.

    Attributes:
      raw_segments: Segments in path (in user space).
      stroke: True if the outline of the path is stroked.
      fill: True if the path is filled.
      evenodd: True if the filling of complex paths uses the even-odd
        winding rule, False if the non-zero winding number rule is
        used (PDF 1.7 section 8.5.3.3)
    """

    raw_segments: List[PathSegment]
    stroke: bool
    fill: bool
    evenodd: bool

    @property
    def segments(self) -> Iterator[PathSegment]:
        """Get path segments in device space."""
        return (
            PathSegment(
                p.operator,
                tuple(apply_matrix_pt(self.ctm, point) for point in p.points),
            )
            for p in self.raw_segments
        )

    @property
    def bbox(self) -> Rect:
        """Get bounding box of path in device space as defined by its
        points and control points."""
        # First get the bounding box in user space (fast)
        bbox = get_bound(
            itertools.chain.from_iterable(seg.points for seg in self.raw_segments)
        )
        # Transform it and get the new bounding box
        return transform_bbox(self.ctm, bbox)


class TextBase(ContentObject):
    """Common properties for text and glyph objects."""

    @property
    @abstractmethod
    def matrix(self) -> Matrix:
        """Rendering matrix `T_rm`, which transforms text space coordinates to
        device space (PDF 2.0 section 9.4.4)."""

    @property
    @abstractmethod
    def displacement(self) -> Point:
        """Vector to the origin of the next glyph in device space."""

    @property
    def font(self) -> Font:
        """Font for this text object."""
        font = self.gstate.font
        assert font is not None
        return font

    @property
    def size(self) -> float:
        """Font size for this text object.

        This is the actual font size in device space, which is **not**
        the same as `GraphicState.fontsize`.  That's the font size in
        text space which is not a very useful number (it's usually 1).
        """
        vert = False if self.gstate.font is None else self.gstate.font.vertical
        if vert:
            # dx, dy = apply_matrix_norm(self.matrix, (1, 0))
            dx, dy, _, _, _, _ = self.matrix
        else:
            # dx, dy = apply_matrix_norm(self.matrix, (0, 1))
            _, _, dx, dy, _, _ = self.matrix
        if dx == 0:  # Nearly always true
            return abs(dy)
        elif dy == 0:
            return abs(dx)
        else:
            import math

            return math.sqrt(dx * dx + dy * dy)

    @property
    def fontname(self) -> str:
        """Font name for this text object"""
        return self.font.fontname

    @property
    def fontbase(self) -> str:
        """Original font name for this text object.

        Fonts in PDF files are usually "subsetted", meaning only the
        glyphs actually used in the document are included.  In this
        case the font's `fontname` property usually consists of an
        arbitrary "tag", plus (literally, a `+`) and the original
        name.  This is a convenience property to get that original
        name.

        This is not the same as `GraphicState.font.basefont` which
        usually also includes the subset tag.

        """
        fontname = self.fontname
        subset, _, base = fontname.partition("+")
        if base:
            return base
        return fontname

    @property
    def textfont(self) -> str:
        """Convenient short form of the font name and size.

        For example, "Helvetica 12".
        """
        return f"{self.fontbase} {round(self.size)}"

    @property
    def origin(self) -> Point:
        """Origin of this text object in device space."""
        _, _, _, _, dx, dy = self.matrix
        return dx, dy


@dataclass
class GlyphObject(TextBase):
    """Individual glyph on the page.

    Attributes:
      font: Font for this glyph.
      size: Effective font size for this glyph.
      fontname: Font name.
      fontbase: Short (non-subset) font name.
      textfont: Combined short name and size for the font.
      cid: Character ID for this glyph.
      text: Unicode mapping of this glyph, if any.
      matrix: Rendering matrix `T_rm` for this glyph, which transforms
              text space coordinates to device space (PDF 2.0 section
              9.4.4).
      origin: Origin of this glyph in device space.
      displacement: Vector to the origin of the next glyph in device space.
      bbox: glyph bounding box in device space.

    """

    cid: int
    text: Union[str, None]
    _matrix: Matrix
    _displacement: float
    _corners: bool

    def __iter__(self) -> Iterator[ContentObject]:
        """Possibly iterate over paths in a glyph.

        For Type3 fonts, you can iterate over paths (or anything
        else) inside a glyph, in the coordinate space defined by the
        text rendering matrix.

        Otherwise, you can't do that, and you get nothing.
        """
        from playa.interp import Type3Interpreter

        font = self.font
        itor: Iterator[ContentObject] = iter(())
        if not isinstance(font, Type3Font):
            return itor
        gid = font.encoding.get(self.cid)
        if gid is None:
            log.warning("Unknown CID %d in Type3 font %r", self.cid, font)
            return itor
        charproc = resolve1(font.charprocs.get(gid))
        if not isinstance(charproc, ContentStream):
            log.warning("CharProc %s not found in font %r ", gid, font)
            return itor

        interp = Type3Interpreter(
            self.page,
            [charproc],
            font.resources,
            ctm=mult_matrix(font.matrix, self.matrix),
            # NOTE: no copy here because an interpreter always creates
            # a new graphics state.
            gstate=self.gstate,
        )
        itor = iter(interp)
        # TODO: We *could* try to get and use the d1 information here
        # but if we do that, we need to do it everywhere the glyph is
        # used so that the bbox will be consistent
        return itor

    @property
    def matrix(self) -> Matrix:
        return self._matrix

    @property
    def chars(self) -> str:
        return self.text or ""

    @property
    def displacement(self) -> Point:
        # Equivalent to:
        # apply_matrix_norm(self.matrix,
        #                   (0, self._displacement)
        #                   if font.vertical else
        #                   (self._displacement, 0))
        a, b, c, d, _, _ = self.matrix
        if self.font.vertical:
            return c * self._displacement, d * self._displacement
        else:
            return a * self._displacement, b * self._displacement

    @property
    def bbox(self) -> Rect:
        x0, y0, x1, y1 = self.font.char_bbox(self.cid)
        if self._corners:
            return get_bound(
                (
                    apply_matrix_pt(self.matrix, (x0, y0)),
                    apply_matrix_pt(self.matrix, (x0, y1)),
                    apply_matrix_pt(self.matrix, (x1, y1)),
                    apply_matrix_pt(self.matrix, (x1, y0)),
                )
            )
        else:
            x0, y0 = apply_matrix_pt(self.matrix, (x0, y0))
            x1, y1 = apply_matrix_pt(self.matrix, (x1, y1))
            if x1 < x0:
                x0, x1 = x1, x0
            if y1 < y0:
                y0, y1 = y1, y0
            return (x0, y0, x1, y1)


@dataclass
class TextObject(TextBase):
    """Text object (contains one or more glyphs).

    Attributes:

      matrix: Initial rendering matrix `T_rm` for this text object,
              which transforms text space coordinates to device space
              (PDF 2.0 section 9.4.4).
      origin: Origin of this text object in device space.
      displacement: Vector to the origin of the next text object in
                    device space.
      font: Font for this text object.
      size: Effective font size for this text object.
      fontname: Font name.
      fontbase: Short (non-subset) font name.
      textfont: Combined short name and size for the font.
      text_matrix: Text matrix `T_m` for this text object, which
                   transforms text space coordinates to user space.
      line_matrix: Text line matrix `T_lm` for this text object, which
                   is the text matrix at the beginning of the "current
                   line" (PDF 2.0 section 9.4.1).  Note that this is
                   **not** reliable for detecting line breaks.
      scaling_matrix: The anonymous but rather important matrix which
                      applies font size, horizontal scaling and rise to
                      obtain the rendering matrix (PDF 2.0 sec 9.4.4).
      args: Strings or position adjustments.
      bbox: Text bounding box in device space.

    """

    args: List[Union[bytes, float]]
    line_matrix: Matrix
    _glyph_offset: Point

    _matrix: Union[Matrix, None] = None
    _chars: Union[List[str], None] = None
    _bbox: Union[Rect, None] = None
    _next_glyph_offset: Union[Point, None] = None
    _displacement: Union[Point, None] = None

    def __iter__(self) -> Iterator[GlyphObject]:
        """Generate glyphs for this text object"""
        glyph_offset = self._glyph_offset
        font = self.gstate.font
        # If no font is set, we cannot do anything, since even calling
        # TJ with a displacement and no text effects requires us at
        # least to know the fontsize.
        if font is None:
            log.warning(
                "No font is set, will not update text state or output text: %r TJ",
                self.args,
            )
            self._next_glyph_offset = glyph_offset
            return
        assert self.ctm is not None

        tlm_ctm = mult_matrix(self.line_matrix, self.ctm)
        # Pre-determine if we need to recompute the bound for rotated glyphs
        a, b, c, d, _, _ = tlm_ctm
        corners = b * d < 0 or a * c < 0
        fontsize = self.gstate.fontsize
        horizontal_scaling = self.gstate.scaling * 0.01
        # PDF 2.0 section 9.3.2: The character-spacing parameter, Tc,
        # shall be a number specified in unscaled text space units
        # (although it shall be subject to scaling by the Th parameter
        # if the writing mode is horizontal).
        if fontsize == 0.0:
            # A particular way of "hiding" text...
            scaled_charspace = scaled_wordspace = 0.0
        else:
            scaled_charspace = self.gstate.charspace / fontsize
            # Section 9.3.3: Word spacing "works the same way"
            scaled_wordspace = self.gstate.wordspace / fontsize

        # PDF 2.0 section 9.4.4: Conceptually, the entire
        # transformation from text space to device space can be
        # represented by a text rendering matrix, T_rm:
        #
        # (scaling_matrix @ text_matrix @ glyph.ctm)
        #
        # Note that scaling_matrix and text_matrix are constant across
        # glyphs in a TextObject, and scaling_matrix is always
        # diagonal (thus the mult_matrix call below can be optimized)
        scaling_matrix = (
            fontsize * horizontal_scaling,
            0,
            0,
            fontsize,
            0,
            self.gstate.rise,
        )
        vert = font.vertical
        # FIXME: THIS IS NOT TRUE!!!  We need a test for it though.
        if font.multibyte:
            scaled_wordspace = 0
        (x, y) = glyph_offset
        pos = y if vert else x
        for obj in self.args:
            if isinstance(obj, (int, float)):
                if vert:
                    pos -= obj * 0.001 * fontsize
                else:
                    pos -= obj * 0.001 * fontsize * horizontal_scaling
            else:
                for cid, text in font.decode(obj):
                    glyph_offset = (x, pos) if vert else (pos, y)
                    disp = font.vdisp(cid) if vert else font.hdisp(cid)
                    disp += scaled_charspace
                    if cid == 32:
                        disp += scaled_wordspace
                    matrix = mult_matrix(
                        scaling_matrix, translate_matrix(tlm_ctm, glyph_offset)
                    )
                    glyph = GlyphObject(
                        _pageref=self._pageref,
                        _parentkey=self._parentkey,
                        gstate=self.gstate,
                        ctm=self.ctm,
                        mcstack=self.mcstack,
                        cid=cid,
                        text=text,
                        _matrix=matrix,
                        _displacement=disp,
                        _corners=corners,
                    )
                    yield glyph
                    # This implements the proper scaling of charspace/wordspace
                    if vert:
                        pos += disp * fontsize
                    else:
                        pos += disp * fontsize * horizontal_scaling
        glyph_offset = (x, pos) if vert else (pos, y)
        if self._next_glyph_offset is None:
            self._next_glyph_offset = glyph_offset

    def _get_next_glyph_offset(self) -> Point:
        """Update only the glyph offset without calculating anything else."""
        if self._next_glyph_offset is not None:
            return self._next_glyph_offset
        self._next_glyph_offset = self._glyph_offset
        font = self.gstate.font
        if font is None:
            log.warning(
                "No font is set, will not update text state or output text: %r TJ",
                self.args,
            )
            return self._next_glyph_offset
        if len(self.args) == 0:
            return self._next_glyph_offset

        self._next_glyph_offset = update_glyph_offset(
            self._glyph_offset, font, self.gstate, self.args
        )
        return self._next_glyph_offset

    @property
    def matrix(self) -> Matrix:
        if self._matrix is not None:
            return self._matrix
        self._matrix = mult_matrix(
            self.scaling_matrix, mult_matrix(self.text_matrix, self.ctm)
        )
        return self._matrix

    @property
    def scaling_matrix(self) -> Matrix:
        horizontal_scaling = self.gstate.scaling * 0.01
        fontsize = self.gstate.fontsize
        return (
            fontsize * horizontal_scaling,
            0,
            0,
            fontsize,
            0,
            self.gstate.rise,
        )

    @property
    def text_matrix(self) -> Matrix:
        return translate_matrix(self.line_matrix, self._glyph_offset)

    @property
    def displacement(self) -> Point:
        if self._displacement is not None:
            return self._displacement
        matrix = self.matrix
        next_matrix = mult_matrix(
            self.scaling_matrix,
            mult_matrix(
                translate_matrix(self.line_matrix, self._get_next_glyph_offset()),
                self.ctm,
            ),
        )
        self._displacement = next_matrix[-2] - matrix[-2], next_matrix[-1] - matrix[-1]
        return self._displacement

    @property
    def bbox(self) -> Union[Rect, None]:
        # We specialize this to avoid it having side effects on the
        # text state (already it's a bit of a footgun that __iter__
        # does that...), but also because we know all glyphs have the
        # same text matrix and thus we can avoid a lot of multiply
        if self._bbox is not None:
            return self._bbox
        font = self.gstate.font
        if font is None:
            log.warning(
                "No font is set, will not update text state or output text: %r TJ",
                self.args,
            )
            self._bbox = None
            self._next_glyph_offset = self._glyph_offset
            return self._bbox
        if len(self.args) == 0:
            self._bbox = None
            self._next_glyph_offset = self._glyph_offset
            return self._bbox

        fontsize = self.gstate.fontsize
        rise = self.gstate.rise
        matrix = mult_matrix(self.line_matrix, self.ctm)
        horizontal_scaling = self.gstate.scaling * 0.01
        charspace = self.gstate.charspace
        wordspace = self.gstate.wordspace
        vert = font.vertical
        if font.multibyte:
            wordspace = 0
        (x, y) = self._glyph_offset
        pos = y if vert else x
        x0 = x1 = x
        y0 = y1 = y + rise
        fast_path = False
        if not vert:
            # Scale charspace and wordspace, PDF 2.0 section 9.3.2
            charspace *= horizontal_scaling
            wordspace *= horizontal_scaling
            # Detect the most frequent case, horizontal writing with
            # diagonal font.matrix
            a, b, c, d, e, f = font.matrix
            if b == 0 and c == 0:
                fast_path = True
                y0 += d * font.descent * fontsize
                y1 += d * font.ascent * fontsize
        for obj in self.args:
            if isinstance(obj, (int, float)):
                pos -= obj * 0.001 * fontsize * horizontal_scaling
            else:
                for cid, _ in font.decode(obj):
                    x, y = (x, pos) if vert else (pos, y)
                    if vert:
                        assert isinstance(font, CIDFont)
                        pos += font.vdisp(cid) * fontsize
                    else:
                        hdisp = font.hdisp(cid)
                        pos += hdisp * fontsize * horizontal_scaling
                    if fast_path:
                        x1 = pos
                    else:
                        gx0, gy0, gx1, gy1 = font.char_bbox(cid)
                        gx0 *= fontsize * horizontal_scaling
                        gx1 *= fontsize * horizontal_scaling
                        gy0 *= fontsize
                        gy0 += rise
                        gy1 *= fontsize
                        gy1 += rise
                        x0 = min(x0, x + gx0)
                        y0 = min(y0, y + gy0)
                        x1 = max(x1, x + gx1)
                        y1 = max(y1, y + gy1)
                    pos += charspace
                    if cid == 32:
                        pos += wordspace
        # Update this because we can!
        if self._next_glyph_offset is None:
            self._next_glyph_offset = (x, pos) if vert else (pos, y)
        self._bbox = transform_bbox(matrix, (x0, y0, x1, y1))
        return self._bbox

    @property
    def chars(self) -> str:
        """Get the Unicode characters (in stream order) for this object."""
        if self._chars is not None:
            return "".join(self._chars)
        self._chars = []
        font = self.gstate.font
        assert font is not None, "No font was selected"
        for obj in self.args:
            if not isinstance(obj, bytes):
                continue
            for _, text in font.decode(obj):
                self._chars.append(text)
        return "".join(self._chars)

    def __len__(self) -> int:
        """Return the number of glyphs that would result from iterating over
        this object.

        Important: this is the number of glyphs, *not* the number of
        Unicode characters.
        """
        nglyphs = 0
        font = self.gstate.font
        assert font is not None, "No font was selected"
        for obj in self.args:
            if not isinstance(obj, bytes):
                continue
            nglyphs += sum(1 for _ in font.decode(obj))
        return nglyphs


class ContentSection(Iterable[ContentObject], Sized):
    """Sequence of content objects in a marked content section.

    This is a `Sized` collection so that you can quickly check if it
    is non-empty by its truth value.  The actual length may or may not
    be relevant.
    """

    def __init__(self, objs: Iterable[ContentObject]) -> None:
        self._objs = [obj.finalize() for obj in objs]
        self._texts: Union[List[str], None] = None

    def __len__(self) -> int:
        return len(self._objs)

    def __iter__(self) -> Iterator[ContentObject]:
        return iter(self._objs)

    @property
    def texts(self) -> Sequence[str]:
        """Sequence of text strings for a marked content section."""
        if self._texts is not None:
            return self._texts
        self._texts = []
        for obj in self._objs:
            if not isinstance(obj, TextObject):
                continue
            mcs = obj.mcs
            if mcs is None or mcs.mcid is None:
                continue
            if "ActualText" in mcs.props:
                assert isinstance(mcs.props["ActualText"], bytes)
                chars = decode_text(mcs.props["ActualText"])
            else:
                chars = obj.chars
            # Remove soft hyphens
            chars = chars.replace("\xad", "")
            self._texts.append(chars)
        return self._texts


class ContentSequence(Sequence[ContentSection]):
    """Collect content object in marked content sections.

    These are ordered by marked content ID, and you can index into
    this to get the content for a specific MCID.

    You can iterate over them in page content order with the
    `page_order` property.

    """

    def __init__(self, streamer: Iterable[ContentObject]) -> None:
        from playa.interp import LazyInterpreter

        self._contents: Dict[int, ContentSection] = {}
        self._maxid: int = 0
        for mcid, objs in itertools.groupby(streamer, operator.attrgetter("mcid")):
            if mcid is None:
                continue
            # Python dicts preserve insertion order, but if there are
            # duplicate marked content sections (this is forbidden by
            # the spec, but.....) we can't do page content order
            self._contents[mcid] = ContentSection(objs)
            self._maxid = max(self._maxid, mcid)
        # In this case we can also know about empty content sections
        # as well as the start and end positions
        if isinstance(streamer, LazyInterpreter):
            for mcid in streamer._mcstart:
                self._maxid = max(self._maxid, mcid)
                if mcid not in self._contents:
                    self._contents[mcid] = ContentSection([])

    def __len__(self) -> int:
        return self._maxid + 1

    @property
    def page_order(self) -> Iterator[ContentSection]:
        """Marked content sections in page content order."""
        yield from self._contents.values()

    @overload
    def __getitem__(self, mcid: int) -> ContentSection: ...

    @overload
    def __getitem__(self, mcid: slice) -> Sequence[ContentSection]: ...

    def __getitem__(
        self, mcid: Union[int, slice]
    ) -> Union[ContentSection, Sequence[ContentSection]]:
        if isinstance(mcid, slice):
            return [self[idx] for idx in range(mcid.start, mcid.stop, mcid.step)]
        else:
            if mcid > self._maxid:
                raise IndexError(f"Marked content ID {mcid} out of range")
            return self._contents.get(mcid, ContentSection([]))



import itertools
from typing import cast

import pytest
from playa import asobj
from playa.document import _format_page_label
from playa.pdftypes import LIT
from playa.utils import (
    IDENTITY_MAPPING,
    Matrix,
    apply_matrix_norm,
    apply_matrix_pt,
    decode_text,
    format_int_alpha,
    format_int_roman,
    get_bound,
    normalize_rect,
    transform_bbox,
)


def test_rotated_bboxes() -> None:
    """Verify that rotated bboxes are correctly calculated."""
    points = ((0, 0), (0, 100), (100, 100), (100, 0))
    bbox = (0, 0, 100, 100)
    # Test all possible sorts of CTM
    vals = (-1, -0.5, 0, 0.5, 1)
    for matrix in itertools.product(vals, repeat=4):
        ctm = cast(Matrix, (*matrix, 0, 0))
        bound = get_bound((apply_matrix_pt(ctm, p) for p in points))
        assert transform_bbox(ctm, bbox) == bound


def test_decode_text() -> None:
    """Make sure we can always decode text, even if it is nonsense."""
    assert (
        decode_text(
            b"\xfe\xffMicrosoft\xae Word 2010; modified using iText 2.1.7 by 1T3XT"
        )
        == "Microsoft\xae Word 2010; modified using iText 2.1.7 by 1T3XT"
    )
    assert decode_text(b"\xff\xfeW\x00T\x00F\x00-\x001\x006\x00") == "WTF-16"
    # Doesn't really belong here but let's test asobj_bytes too
    assert asobj(
        b"\xfe\xffMicrosoft\xae Word 2010; modified using iText 2.1.7 by 1T3XT"
    ) == (
        "base64:/v9NaWNyb3NvZnSuIFdvcmQgMjAxMDsgbW9kaWZpZWQgdXNpbmcgaVRleHQgMi4xLj"
        "cgYnkgMVQzWFQ="
    )
    assert asobj(b"\xff\xfeW\x00T\x00F\x00-\x001\x006\x00") == "WTF-16"


def test_format_romans() -> None:
    """Chic, des Romains."""
    with pytest.raises(ValueError):
        format_int_roman(4000)
    with pytest.raises(ValueError):
        format_int_roman(-1)
    with pytest.raises(ValueError):
        format_int_roman(0)

    romans = ["", *(format_int_roman(x) for x in range(1, 4000))]
    assert romans[9] == "ix"
    for x in range(10, 50):
        assert romans[x].startswith("x")
    for x in range(50, 90):
        assert romans[x].startswith("l")
    for x in range(90, 100):
        assert romans[x].startswith("xc")
    for x in range(100, 500):
        assert romans[x].startswith("c")
    for x in range(500, 900):
        assert romans[x].startswith("d")
    for x in range(900, 1000):
        assert romans[x].startswith("cm")

    romans2 = [
        "",
        *(_format_page_label(x, LIT("r")) for x in range(1, 4000)),
    ]
    assert romans2 == romans

    ROMANS = ["", *(_format_page_label(x, LIT("R")) for x in range(1, 4000))]
    assert ROMANS == [x.upper() for x in romans]


def test_format_alphas() -> None:
    """Vendeurs de thermopompes."""

    with pytest.raises(ValueError):
        format_int_alpha(0)

    two_letters = (26 + 1) * 26
    alphas = ["", *(format_int_alpha(x) for x in range(1, two_letters + 1))]
    assert alphas[-1] == "zz"
    assert format_int_alpha(two_letters + 1) == "aaa"

    alphas2 = [
        "",
        *(_format_page_label(x, LIT("a")) for x in range(1, two_letters + 1)),
    ]
    assert alphas2 == alphas

    ALPHAS = [
        "",
        *(_format_page_label(x, LIT("A")) for x in range(1, two_letters + 1)),
    ]
    assert ALPHAS == [x.upper() for x in alphas]


def test_normalize_rect() -> None:
    """Normalize rects"""
    r1 = (1, 1, 5, 5)
    assert normalize_rect(r1) == r1
    r2 = (5, 5, 1, 1)
    assert normalize_rect(r2) == r1


def test_apply_matrix_norm() -> None:
    m = (1, 0.75, -0.75, 1, 3, 4)
    x, y = (123, 456)
    nx, ny = apply_matrix_norm(m, (x, y))
    zx, zy = apply_matrix_pt(m, (0, 0))
    px, py = apply_matrix_pt(m, (x, y))
    assert nx == px - zx
    assert ny == py - zy


def test_identity_mapping() -> None:
    assert IDENTITY_MAPPING[42] == 42
    assert IDENTITY_MAPPING["xviii"] == "xviii"



"""Python implementation of ASCII85/ASCIIHex decoder (Adobe version)."""

import re
from base64 import a85decode
from binascii import unhexlify

start_re = re.compile(rb"^\s*<?\s*~\s*")
end_re = re.compile(rb"\s*~\s*>?\s*$")


def ascii85decode(data: bytes) -> bytes:
    """In ASCII85 encoding, every four bytes are encoded with five ASCII
    letters, using 85 different types of characters (as 256**4 < 85**5).
    When the length of the original bytes is not a multiple of 4, a special
    rule is used for round up.

    Adobe's ASCII85 implementation expects the input to be terminated
    by `b"~>"`, and (though this is absent from the PDF spec) it can
    also begin with `b"<~"`.  We can't reliably expect this to be the
    case, and there can be off-by-one errors in stream lengths which
    mean we only see `~` at the end.  Worse yet, `<` and `>` are
    ASCII85 digits, so we can't strip them.  We settle on a compromise
    where we strip leading `<~` or `~` and trailing `~` or `~>`.
    """
    data = start_re.sub(b"", data)
    data = end_re.sub(b"", data)
    return a85decode(data)


bws_re = re.compile(rb"\s")


def asciihexdecode(data: bytes) -> bytes:
    """ASCIIHexDecode filter: PDFReference v1.4 section 3.3.1
    For each pair of ASCII hexadecimal digits (0-9 and A-F or a-f), the
    ASCIIHexDecode filter produces one byte of binary data. All white-space
    characters are ignored. A right angle bracket character (>) indicates
    EOD. Any other characters will cause an error. If the filter encounters
    the EOD marker after reading an odd number of hexadecimal digits, it
    will behave as if a 0 followed the last digit.
    """
    data = bws_re.sub(b"", data)
    end = data.find(b">")
    if end != -1:
        data = data[:end]
    if len(data) % 2 == 1:
        data += b"0"
    return unhexlify(data)



from playa.ccitt import CCITTFaxDecoder, CCITTG4Parser


class TestCCITTG4Parser:
    def get_parser(self, bits):
        parser = CCITTG4Parser(len(bits), 2)
        parser._curline = [int(c) for c in bits]
        parser._reset_line()
        return parser

    def test_b1(self):
        parser = self.get_parser("00000")
        parser._do_vertical(0)
        assert parser._curpos == 0

    def test_b2(self):
        parser = self.get_parser("10000")
        parser._do_vertical(-1)
        assert parser._curpos == 0

    def test_b3(self):
        parser = self.get_parser("000111")
        parser._do_pass()
        assert parser._curpos == 3
        assert parser._get_bits() == "111"

    def test_b4(self):
        parser = self.get_parser("00000")
        parser._do_vertical(+2)
        assert parser._curpos == 2
        assert parser._get_bits() == "11"

    def test_b5(self):
        parser = self.get_parser("11111111100")
        parser._do_horizontal(0, 3)
        assert parser._curpos == 3
        parser._do_vertical(1)
        assert parser._curpos == 10
        assert parser._get_bits() == "0001111111"

    def test_e1(self):
        parser = self.get_parser("10000")
        parser._do_vertical(0)
        assert parser._curpos == 1
        parser._do_vertical(0)
        assert parser._curpos == 5
        assert parser._get_bits() == "10000"

    def test_e2(self):
        parser = self.get_parser("10011")
        parser._do_vertical(0)
        assert parser._curpos == 1
        parser._do_vertical(2)
        assert parser._curpos == 5
        assert parser._get_bits() == "10000"

    def test_e3(self):
        parser = self.get_parser("011111")
        parser._color = 0
        parser._do_vertical(0)
        assert parser._color == 1
        assert parser._curpos == 1
        parser._do_vertical(-2)
        assert parser._color == 0
        assert parser._curpos == 4
        parser._do_vertical(0)
        assert parser._curpos == 6
        assert parser._get_bits() == "011100"

    def test_e4(self):
        parser = self.get_parser("10000")
        parser._do_vertical(0)
        assert parser._curpos == 1
        parser._do_vertical(-2)
        assert parser._curpos == 3
        parser._do_vertical(0)
        assert parser._curpos == 5
        assert parser._get_bits() == "10011"

    def test_e5(self):
        parser = self.get_parser("011000")
        parser._color = 0
        parser._do_vertical(0)
        assert parser._curpos == 1
        parser._do_vertical(3)
        assert parser._curpos == 6
        assert parser._get_bits() == "011111"

    def test_e6(self):
        parser = self.get_parser("11001")
        parser._do_pass()
        assert parser._curpos == 4
        parser._do_vertical(0)
        assert parser._curpos == 5
        assert parser._get_bits() == "11111"

    def test_e7(self):
        parser = self.get_parser("0000000000")
        parser._curpos = 2
        parser._color = 1
        parser._do_horizontal(2, 6)
        assert parser._curpos == 10
        assert parser._get_bits() == "1111000000"

    def test_e8(self):
        parser = self.get_parser("001100000")
        parser._curpos = 1
        parser._color = 0
        parser._do_vertical(0)
        assert parser._curpos == 2
        parser._do_horizontal(7, 0)
        assert parser._curpos == 9
        assert parser._get_bits() == "101111111"

    def test_m1(self):
        parser = self.get_parser("10101")
        parser._do_pass()
        assert parser._curpos == 2
        parser._do_pass()
        assert parser._curpos == 4
        assert parser._get_bits() == "1111"

    def test_m2(self):
        parser = self.get_parser("101011")
        parser._do_vertical(-1)
        parser._do_vertical(-1)
        parser._do_vertical(1)
        parser._do_horizontal(1, 1)
        assert parser._get_bits() == "011101"

    def test_m3(self):
        parser = self.get_parser("10111011")
        parser._do_vertical(-1)
        parser._do_pass()
        parser._do_vertical(1)
        parser._do_vertical(1)
        assert parser._get_bits() == "00000001"


class TestCCITTFaxDecoder:
    def test_b1(self):
        decoder = CCITTFaxDecoder({"Columns": 5})
        decoder.output_line(0, [1, 0, 0, 0, 1, 0, 1])
        assert decoder.close() == b"\x8a"



"""
Benchmark CCITT decoding.
"""

import logging
import time
from pathlib import Path

import playa

CONTRIB = Path(__file__).parent.parent / "samples" / "contrib"

LOG = logging.getLogger(Path(__file__).stem)
PDFS = ["ccitt-default-k.pdf", "ccitt_EndOfBlock_false.pdf"]


def benchmark_images(path: Path):
    with playa.open(path) as pdf:
        for page in pdf.pages:
            for image in page.images:
                _ = image.buffer


if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    niter = 1
    t = 0.0
    for iter in range(niter + 1):
        for name in PDFS:
            path = CONTRIB / name
            start = time.time()
            benchmark_images(path)
            if iter != 0:
                t += time.time() - start
    print("ccitt took %d ms / iter" % (t / niter * 1000,))



"""
Test the data API
"""

import playa
from dataclasses import dataclass
from typing import NamedTuple


def test_asobj_generic() -> None:
    """Test asobj on generic dataclasses and namedtuples"""

    @dataclass
    class Spam:
        spam: str

    spam = Spam(spam="spam")
    assert playa.asobj(spam) == {"spam": "spam"}

    class Eggs(NamedTuple):
        spam: str

    eggs = Eggs(spam="spam")
    assert playa.asobj(eggs) == {"spam": "spam"}



"""Adobe character mapping (CMap) support.

CMaps provide the mapping between character codes and Unicode
code-points to character ids (CIDs).

More information is available on:

  https://github.com/adobe-type-tools/cmap-resources

"""

import gzip
import logging
import pickle as pickle
import struct
import sys
from bisect import bisect_left
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    TextIO,
    Tuple,
    Union,
)

from playa.exceptions import PDFSyntaxError
from playa.parser import (
    KWD,
    ObjectParser,
    PDFObject,
    PSKeyword,
    literal_name,
)
from playa.utils import choplist

from functools import lru_cache


log = logging.getLogger(__name__)
CMAP_DIR = (Path(__file__).parent / "cmap").resolve()


class CMapError(Exception):
    pass


class CMapBase:
    debug = 0

    def __init__(self, **kwargs: Any) -> None:
        self.attrs: Dict[str, Any] = kwargs.copy()

    def is_vertical(self) -> bool:
        return self.attrs.get("WMode", 0) != 0

    def set_attr(self, k: str, v: Any) -> None:
        self.attrs[k] = v

    def use_cmap(self, cmap: "CMapBase") -> None:
        pass

    def decode(self, code: bytes) -> Iterable[Tuple[bytes, int]]:
        raise NotImplementedError


# The CID map is a sort of trie
CodeToCIDMap = Dict[int, Union[int, "CodeToCIDMap"]]


class CMap(CMapBase):
    def __init__(self, **kwargs: Union[str, int]) -> None:
        CMapBase.__init__(self, **kwargs)
        self.code2cid: CodeToCIDMap = {}

    def __repr__(self) -> str:
        return "<CMap: %s>" % self.attrs.get("CMapName")

    def use_cmap(self, cmap: CMapBase) -> None:
        assert isinstance(cmap, CMap), str(type(cmap))

        def copy(dst: CodeToCIDMap, src: CodeToCIDMap) -> None:
            for k, v in src.items():
                if isinstance(v, int):
                    dst[k] = v
                else:
                    d: CodeToCIDMap = {}
                    dst[k] = d
                    copy(d, v)

        copy(self.code2cid, cmap.code2cid)

    def decode(self, code: bytes) -> Iterator[Tuple[bytes, int]]:
        d = self.code2cid
        substr = []
        for i in iter(code):
            if i in d:
                x = d[i]
                substr.append(i)
                if isinstance(x, int):
                    yield bytes(substr), x
                    d = self.code2cid
                    del substr[:]
                else:
                    d = x
            else:
                d = self.code2cid
                del substr[:]

    def dump(
        self,
        out: TextIO = sys.stdout,
        code2cid: Union[CodeToCIDMap, None] = None,
        code: Tuple[int, ...] = (),
    ) -> None:
        if code2cid is None:
            code2cid = self.code2cid
            code = ()
        for k, v in sorted(code2cid.items()):
            c = code + (k,)
            if isinstance(v, int):
                out.write("code %r = cid %d\n" % (c, v))
            else:
                self.dump(out=out, code2cid=v, code=c)


class IdentityCMap(CMapBase):
    def decode(self, code: bytes) -> Iterable[Tuple[bytes, int]]:
        n = len(code) // 2
        if n:
            codes = (code[x : x + 2] for x in range(0, len(code), 2))
            cids = struct.unpack(">%dH" % n, code)
            return zip(codes, cids)
        else:
            return ()


class IdentityCMapByte(IdentityCMap):
    def decode(self, code: bytes) -> Iterable[Tuple[bytes, int]]:
        n = len(code)
        if n:
            codes = (code[x : x + 1] for x in range(n))
            cids = struct.unpack(">%dB" % n, code)
            return zip(codes, cids)
        else:
            return ()


class UnicodeMap(CMapBase):
    def __init__(self, **kwargs: Union[str, int]) -> None:
        CMapBase.__init__(self, **kwargs)
        self.cid2unichr: Dict[int, str] = {}

    def __repr__(self) -> str:
        return "<UnicodeMap: %s>" % self.attrs.get("CMapName")

    def get_unichr(self, cid: int) -> str:
        if cid not in self.cid2unichr:
            log.warning("CID %d not found in Unicode map %s", cid, self)
            return ""
        return self.cid2unichr[cid]

    def dump(self, out: TextIO = sys.stdout) -> None:
        for k, v in sorted(self.cid2unichr.items()):
            out.write("cid %d = unicode %r\n" % (k, v))


class PyCMap(CMap):
    def __init__(self, name: str, data: Dict) -> None:
        super().__init__(CMapName=name)
        self.code2cid = data["CODE2CID"]
        if data["IS_VERTICAL"]:
            self.attrs["WMode"] = 1


class PyUnicodeMap(UnicodeMap):
    def __init__(self, name: str, data: Dict, vertical: bool) -> None:
        super().__init__(CMapName=name)
        if vertical:
            self.cid2unichr = data["CID2UNICHR_V"]
            self.attrs["WMode"] = 1
        else:
            self.cid2unichr = data["CID2UNICHR_H"]


CMAP_CACHE: Dict[str, PyCMap] = {}
UMAP_CACHE: Dict[str, List[PyUnicodeMap]] = {}


def _load_data(name: str) -> Any:
    name = name.replace("\0", "")
    filename = "%s.pickle.gz" % name
    pklpath = (CMAP_DIR / filename).resolve()
    try:
        _ = pklpath.relative_to(CMAP_DIR)
        with gzip.open(pklpath) as gzfile:
            return pickle.load(gzfile)
    except ValueError as e:
        raise KeyError(f"Ignoring malicious or malformed CMap {name}") from e
    except FileNotFoundError as e:
        raise KeyError(f"CMap {name} not found in CMapDB") from e


def get_cmap(name: str) -> CMapBase:
    if name == "Identity-H":
        return IdentityCMap(CMapName=name, WMode=0)
    elif name == "Identity-V":
        return IdentityCMap(CMapName=name, WMode=1)
    elif name == "OneByteIdentityH":
        return IdentityCMapByte(CMapName=name, WMode=0)
    elif name == "OneByteIdentityV":
        return IdentityCMapByte(CMapName=name, WMode=1)
    if name in CMAP_CACHE:
        return CMAP_CACHE[name]
    data = _load_data(name)
    CMAP_CACHE[name] = cmap = PyCMap(name, data)
    return cmap


def get_unicode_map(name: str, vertical: bool = False) -> UnicodeMap:
    try:
        return UMAP_CACHE[name][vertical]
    except KeyError:
        pass
    data = _load_data("to-unicode-%s" % name)
    UMAP_CACHE[name] = [PyUnicodeMap(name, data, v) for v in (False, True)]
    return UMAP_CACHE[name][vertical]


KEYWORD_BEGINCMAP = KWD(b"begincmap")
KEYWORD_ENDCMAP = KWD(b"endcmap")
KEYWORD_USECMAP = KWD(b"usecmap")
KEYWORD_DEF = KWD(b"def")
KEYWORD_BEGINCODESPACERANGE = KWD(b"begincodespacerange")
KEYWORD_ENDCODESPACERANGE = KWD(b"endcodespacerange")
KEYWORD_BEGINCIDRANGE = KWD(b"begincidrange")
KEYWORD_ENDCIDRANGE = KWD(b"endcidrange")
KEYWORD_BEGINCIDCHAR = KWD(b"begincidchar")
KEYWORD_ENDCIDCHAR = KWD(b"endcidchar")
KEYWORD_BEGINBFRANGE = KWD(b"beginbfrange")
KEYWORD_ENDBFRANGE = KWD(b"endbfrange")
KEYWORD_BEGINBFCHAR = KWD(b"beginbfchar")
KEYWORD_ENDBFCHAR = KWD(b"endbfchar")
KEYWORD_BEGINNOTDEFRANGE = KWD(b"beginnotdefrange")
KEYWORD_ENDNOTDEFRANGE = KWD(b"endnotdefrange")


# These are generally characters or short strings (glyph clusters) so
# caching them infinitely makes sense (they repeat themselves often)
@lru_cache(None)
def decode_utf16_char(utf16: bytes) -> str:
    return utf16.decode("UTF-16BE", "ignore")


class ToUnicodeMap:
    """ToUnicode map loaded from a PDF stream.  Not a CMap!"""

    def __init__(self) -> None:
        self.attrs: Dict[str, Any] = {}
        self.bytes2unicode: Dict[bytes, str] = {}
        self.code_lengths: List[int] = []
        self.code_space: List[Tuple[bytes, bytes]] = []

    def set_attr(self, k: str, v: Any) -> None:
        self.attrs[k] = v

    def use_cmap(self, cmap: "CMapBase") -> None:
        # FIXME: This should probably do ... something?
        pass

    def add_code_range(self, start: bytes, end: bytes):
        """Add a code-space range"""
        if len(start) != len(end):
            log.warning(
                "Ignoring inconsistent code lengths in code space: %r / %r", start, end
            )
            return
        codelen = len(start)
        pos = bisect_left(self.code_lengths, codelen)
        self.code_lengths.insert(pos, codelen)
        self.code_space.insert(pos, (start, end))

    def decode(self, code: bytes) -> Iterator[str]:
        """Decode a multi-byte string to Unicode sequences"""
        idx = 0
        codelen = 1
        while idx < len(code):
            # Match code space ranges
            for codelen, (start, end) in zip(self.code_lengths, self.code_space):
                substr = code[idx : idx + codelen]
                # NOTE: lexicographical ordering is the same as
                # big-endian numerical ordering so this works
                if substr >= start and substr <= end:
                    if substr in self.bytes2unicode:
                        yield self.bytes2unicode[substr]
                    else:
                        log.debug("Undefined character code %r", substr)
                        yield chr(int.from_bytes(substr, "big"))
                    idx += codelen
                    break
            else:
                log.debug("No code space found for %r", code[idx:])
                yield chr(code[idx])
                idx += 1

    def add_code2bytes(self, code: int, utf16: bytes, codelen: int) -> None:
        self.bytes2unicode[code.to_bytes(codelen, "big")] = decode_utf16_char(utf16)

    def add_code2code(self, code: int, outcode: int, codelen: int) -> None:
        uni = chr(outcode)
        self.bytes2unicode[code.to_bytes(codelen, "big")] = uni

    def add_bf_range(self, start_byte: bytes, end_byte: bytes, code: PDFObject) -> None:
        start = int.from_bytes(start_byte, "big")
        end = int.from_bytes(end_byte, "big")
        codelen = len(start_byte)
        if isinstance(code, list):
            if len(code) != end - start + 1:
                log.warning(
                    "The difference between the start and end "
                    "offsets does not match the code length.",
                )
            for cid, unicode_value in zip(range(start, end + 1), code):
                assert isinstance(unicode_value, bytes)
                self.add_code2bytes(cid, unicode_value, codelen)
        elif isinstance(code, bytes):
            unibase = int.from_bytes(code, "big")
            unilen = len(code)
            for i in range(end - start + 1):
                self.add_code2bytes(
                    start + i, (unibase + i).to_bytes(unilen, "big"), codelen
                )
        elif isinstance(code, int):
            for i in range(end - start + 1):
                self.add_code2code(start + i, code + i, codelen)
        else:
            raise ValueError("Unuspported character code %r", code)


def parse_tounicode(data: bytes) -> ToUnicodeMap:
    cmap = ToUnicodeMap()
    stack: List[PDFObject] = []
    parser = ObjectParser(data)
    # some ToUnicode maps don't have "begincmap" keyword.
    in_cmap = True

    while True:
        try:
            pos, obj = next(parser)
        except PDFSyntaxError:
            # CMap syntax is apparently not PDF syntax (e.g. "def"
            # seems to occur within dictionaries, for no apparent
            # reason, perhaps a PostScript thing?)
            parser.reset()
            continue
        except StopIteration:
            break

        if not isinstance(obj, PSKeyword):
            stack.append(obj)
            continue
        # Ignore everything outside begincmap / endcmap
        if obj is KEYWORD_BEGINCMAP:
            in_cmap = True
            del stack[:]
        elif obj is KEYWORD_ENDCMAP:
            in_cmap = False
        if not in_cmap:
            return cmap

        if obj is KEYWORD_DEF:
            try:
                # Might fail with IndexError if the file is corrputed
                v = stack.pop()
                k = stack.pop()
                cmap.set_attr(literal_name(k), v)
            except (IndexError, TypeError):
                pass
        elif obj is KEYWORD_USECMAP:
            try:
                cmapname = stack.pop()
                cmap.use_cmap(get_cmap(literal_name(cmapname)))
            except (IndexError, TypeError, KeyError):
                pass
        elif obj is KEYWORD_BEGINCODESPACERANGE:
            del stack[:]
        elif obj is KEYWORD_ENDCODESPACERANGE:
            for start_code, end_code in choplist(2, stack):
                if not isinstance(start_code, bytes):
                    log.warning(
                        "Start of code space range %r %r is not bytes.",
                        start_code,
                        end_code,
                    )
                    return cmap
                if not isinstance(end_code, bytes):
                    log.warning(
                        "End of code space range %r %r is not bytes.",
                        start_code,
                        end_code,
                    )
                    return cmap
                cmap.add_code_range(start_code, end_code)
            del stack[:]
        elif obj is KEYWORD_BEGINCIDRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDCIDRANGE:
            for start_byte, end_byte, code in choplist(3, stack):
                if not isinstance(start_byte, bytes):
                    log.warning("The start object is not a byte.")
                    continue
                if not isinstance(end_byte, bytes):
                    log.warning("The end object is not a byte.")
                    continue
                if len(start_byte) != len(end_byte):
                    log.warning("The start and end byte have different lengths.")
                    continue
                cmap.add_bf_range(start_byte, end_byte, code)
            del stack[:]
        elif obj is KEYWORD_BEGINCIDCHAR:
            del stack[:]
        elif obj is KEYWORD_ENDCIDCHAR:
            for cid, code in choplist(2, stack):
                if isinstance(cid, bytes) and isinstance(code, int):
                    cmap.add_code2code(int.from_bytes(cid, "big"), code, len(cid))
            del stack[:]
        elif obj is KEYWORD_BEGINBFRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDBFRANGE:
            for start_byte, end_byte, code in choplist(3, stack):
                if not isinstance(start_byte, bytes):
                    log.warning("The start object is not a byte.")
                    continue
                if not isinstance(end_byte, bytes):
                    log.warning("The end object is not a byte.")
                    continue
                if len(start_byte) != len(end_byte):
                    log.warning("The start and end byte have different lengths.")
                    continue
                cmap.add_bf_range(start_byte, end_byte, code)
            del stack[:]
        elif obj is KEYWORD_BEGINBFCHAR:
            del stack[:]
        elif obj is KEYWORD_ENDBFCHAR:
            for cid, code in choplist(2, stack):
                if isinstance(cid, bytes) and isinstance(code, bytes):
                    cmap.add_code2bytes(int.from_bytes(cid, "big"), code, len(cid))
            del stack[:]
        elif obj is KEYWORD_BEGINNOTDEFRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDNOTDEFRANGE:
            del stack[:]
        else:
            # It's ... something else (probably bogus)
            stack.append(obj)
    return cmap


class EncodingCMap(CMap):
    """Encoding map loaded from a PDF stream."""

    def __init__(self) -> None:
        super().__init__()
        self.bytes2cid: Dict[bytes, int] = {}
        self.code_lengths: List[int] = []
        self.code_space: List[Tuple[bytes, bytes]] = []

    def add_code_range(self, start: bytes, end: bytes):
        """Add a code-space range"""
        if len(start) != len(end):
            log.warning(
                "Ignoring inconsistent code lengths in code space: %r / %r", start, end
            )
            return
        codelen = len(start)
        pos = bisect_left(self.code_lengths, codelen)
        self.code_lengths.insert(pos, codelen)
        self.code_space.insert(pos, (start, end))

    def decode(self, code: bytes) -> Iterator[Tuple[bytes, int]]:
        """Decode a multi-byte string according to the CMap"""
        idx = 0
        codelen = 1
        while idx < len(code):
            # Match code space ranges
            for codelen, (start, end) in zip(self.code_lengths, self.code_space):
                substr = code[idx : idx + codelen]
                # NOTE: lexicographical ordering is the same as
                # big-endian numerical ordering so this works
                if substr >= start and substr <= end:
                    if substr not in self.bytes2cid:
                        # 9.7.6.3: If a code maps to a CID for which
                        # no such glyph exists in the descendant
                        # CIDFont...
                        # FIXME: Implement notdef mappings
                        yield substr, 0
                    else:
                        yield substr, self.bytes2cid[substr]
                    idx += codelen
                    break
            else:
                # 9.7.6.3 If the code is invalid—that is, the bytes
                # extracted from the string to be shown do not match
                # any codespace range in the CMap...
                log.debug("No code space found for %r", code[idx:])
                # FIXME: Implement the somewhat obscure partial
                # matching algorithm (might consume more than 1 byte)
                yield code[idx : idx + 1], 0
                idx += 1

    def add_bytes2cid(self, code: bytes, cid: int) -> None:
        self.bytes2cid[code] = cid

    def add_cid_range(self, start_byte: bytes, end_byte: bytes, cid: int) -> None:
        start = int.from_bytes(start_byte, "big")
        end = int.from_bytes(end_byte, "big")
        codelen = len(start_byte)
        for i in range(end - start + 1):
            x = (start + i).to_bytes(codelen, "big")
            self.add_bytes2cid(x, cid + i)


def parse_encoding(data: bytes) -> EncodingCMap:
    """Parse an Encoding CMap."""
    cmap = EncodingCMap()
    stack: List[PDFObject] = []
    parser = ObjectParser(data)

    while True:
        try:
            pos, obj = next(parser)
        except PDFSyntaxError:
            parser.reset()
            continue
        except StopIteration:
            break

        if not isinstance(obj, PSKeyword):
            stack.append(obj)
            continue

        if obj is KEYWORD_DEF:
            try:
                # Might fail with IndexError if the file is corrputed
                v = stack.pop()
                k = stack.pop()
                cmap.set_attr(literal_name(k), v)
            except (IndexError, TypeError):
                pass
        elif obj is KEYWORD_USECMAP:
            log.warning("usecmap not supported for EncodingCMap")
            del stack[:]
        elif obj is KEYWORD_BEGINCODESPACERANGE:
            del stack[:]
        elif obj is KEYWORD_ENDCODESPACERANGE:
            for start_code, end_code in choplist(2, stack):
                if not isinstance(start_code, bytes):
                    log.warning(
                        "Start of code space range %r %r is not bytes.",
                        start_code,
                        end_code,
                    )
                    return cmap
                if not isinstance(end_code, bytes):
                    log.warning(
                        "End of code space range %r %r is not bytes.",
                        start_code,
                        end_code,
                    )
                    return cmap
                cmap.add_code_range(start_code, end_code)
            del stack[:]
        elif obj is KEYWORD_BEGINCIDRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDCIDRANGE:
            for start_byte, end_byte, cid in choplist(3, stack):
                if not isinstance(start_byte, bytes):
                    log.warning("The start object of begincidrange is not a byte.")
                    return cmap
                if not isinstance(end_byte, bytes):
                    log.warning("The end object of begincidrange is not a byte.")
                    return cmap
                if not isinstance(cid, int):
                    log.warning("The cid object of begincidrange is not a byte.")
                    return cmap
                if len(start_byte) != len(end_byte):
                    log.warning(
                        "The start and end byte of begincidrange have different lengths.",
                    )
                    return cmap
                cmap.add_cid_range(start_byte, end_byte, cid)
            del stack[:]
        elif obj is KEYWORD_BEGINCIDCHAR:
            del stack[:]
        elif obj is KEYWORD_ENDCIDCHAR:
            for code, cid in choplist(2, stack):
                if isinstance(code, bytes) and isinstance(cid, int):
                    cmap.add_bytes2cid(code, cid)
            del stack[:]
        elif obj is KEYWORD_BEGINBFRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDBFRANGE:
            del stack[:]
        elif obj is KEYWORD_BEGINBFCHAR:
            del stack[:]
        elif obj is KEYWORD_ENDBFCHAR:
            del stack[:]
        elif obj is KEYWORD_BEGINNOTDEFRANGE:
            del stack[:]
        elif obj is KEYWORD_ENDNOTDEFRANGE:
            del stack[:]
        else:
            # It's ... something else (probably bogus)
            stack.append(obj)
    return cmap



import logging
import functools
import itertools
import struct
from pathlib import Path
from typing import BinaryIO, Callable, Final, Tuple

from playa import asobj
from playa.color import get_colorspace
from playa.pdftypes import (
    LITERALS_DCT_DECODE,
    LITERALS_JPX_DECODE,
    LITERALS_JBIG2_DECODE,
    ContentStream,
    resolve1,
    stream_value,
)

LOG: Final = logging.getLogger(__name__)

JBIG2_HEADER: Final = b"\x97JB2\r\n\x1a\n"


# PDF 2.0, sec 8.9.3 Sample data shall be represented as a stream of
# bytes, interpreted as 8-bit unsigned integers in the range 0 to
# 255. The bytes constitute a continuous bit stream, with the
# high-order bit of each byte first.  This bit stream, in turn, is
# divided into units of n bits each, where n is the number of bits per
# component.  Each unit encodes a colour component value, given with
# high-order bit first; units of 16 bits shall be given with the most
# significant byte first. Byte boundaries shall be ignored, except
# that each row of sample data shall begin on a byte boundary. If the
# number of data bits per row is not a multiple of 8, the end of the
# row is padded with extra bits to fill out the last byte. A PDF
# processor shall ignore these padding bits.
def unpack_image_data(
    s: bytes, bpc: int, width: int, height: int, ncomponents: int
) -> bytes:
    if bpc not in (1, 2, 4):
        return s
    if bpc == 4:

        def unpack_f(x: int) -> Tuple[int, ...]:
            return (x >> 4, x & 15)

    elif bpc == 2:

        def unpack_f(x: int) -> Tuple[int, ...]:
            return (x >> 6, x >> 4 & 3, x >> 2 & 3, x & 3)

    else:  # bpc == 1

        def unpack_f(x: int) -> Tuple[int, ...]:
            return tuple(x >> i & 1 for i in reversed(range(8)))

    rowsize = (width * ncomponents * bpc + 7) // 8
    rows = (s[i * rowsize : (i + 1) * rowsize] for i in range(height))
    unpacked_rows = (
        itertools.islice(
            itertools.chain.from_iterable(map(unpack_f, row)), width * ncomponents
        )
        for row in rows
    )
    return bytes(itertools.chain.from_iterable(unpacked_rows))


def get_one_image(stream: ContentStream, path: Path) -> Path:
    suffix, writer = get_image_suffix_and_writer(stream)
    path = path.with_suffix(suffix)
    with open(path, "wb") as outfh:
        writer(outfh)
    return path


def get_image_suffix_and_writer(
    stream: ContentStream,
) -> Tuple[str, Callable[[BinaryIO], None]]:
    for f, parms in stream.get_filters():
        if f in LITERALS_DCT_DECODE:
            # DCT streams are generally readable as JPEG files
            return ".jpg", functools.partial(write_raw, data=stream.buffer)
        if f in LITERALS_JPX_DECODE:
            # This is also generally true for JPEG2000 streams
            return ".jp2", functools.partial(write_raw, data=stream.buffer)
        if f in LITERALS_JBIG2_DECODE:
            # This is not however true for JBIG2, which requires a
            # particular header
            globals_stream = resolve1(parms.get("JBIG2Globals"))
            if isinstance(globals_stream, ContentStream):
                jbig2globals = globals_stream.buffer
            else:
                jbig2globals = b""
            return ".jb2", functools.partial(
                write_jbig2, data=stream.buffer, jbig2globals=jbig2globals
            )

    bits = stream.bits
    width = stream.width
    height = stream.height
    colorspace = stream.colorspace
    ncomponents = colorspace.ncomponents
    data = stream.buffer
    if bits == 1 and ncomponents == 1 and colorspace.name != "Indexed":
        return ".pbm", functools.partial(
            write_pbm, data=data, width=width, height=height
        )

    data = unpack_image_data(data, bits, width, height, ncomponents)
    # TODO: Decode array goes here
    if colorspace.name == "Indexed":
        assert isinstance(colorspace.spec, list)
        _, underlying, hival, lookup = colorspace.spec
        underlying_colorspace = get_colorspace(resolve1(underlying))
        if underlying_colorspace is None:
            LOG.warning(
                "Unknown underlying colorspace in Indexed image: %r, writing as grayscale",
                resolve1(underlying),
            )
        else:
            ncomponents = underlying_colorspace.ncomponents
            if not isinstance(lookup, bytes):
                lookup = stream_value(lookup).buffer
            data = bytes(
                b for i in data for b in lookup[ncomponents * i : ncomponents * (i + 1)]
            )
            bits = 8

    if ncomponents == 1:
        return ".pgm", functools.partial(
            write_pnm,
            data=data,
            ftype=b"P5",
            bits=bits,
            width=width,
            height=height,
        )
    elif ncomponents == 3:
        return ".ppm", functools.partial(
            write_pnm,
            data=data,
            ftype=b"P6",
            bits=bits,
            width=width,
            height=height,
        )
    elif ncomponents == 4:
        return ".tif", functools.partial(
            write_cmyk_tiff,
            data=data,
            bits=bits,
            width=width,
            height=height,
        )
    else:
        LOG.warning(
            "Unsupported colorspace %r, writing as raw bytes", asobj(colorspace)
        )
        return ".dat", functools.partial(write_raw, data=data)


def write_raw(outfh: BinaryIO, data: bytes) -> None:
    outfh.write(data)


def write_pbm(outfh: BinaryIO, data: bytes, width: int, height: int) -> None:
    """Write stream data to a PBM file."""
    outfh.write(b"P4 %d %d\n" % (width, height))
    outfh.write(bytes(x ^ 0xFF for x in data))


def write_pnm(
    outfh: BinaryIO, data: bytes, ftype: bytes, bits: int, width: int, height: int
) -> None:
    """Write stream data to a PGM/PPM file."""
    max_value = (1 << bits) - 1
    outfh.write(b"%s %d %d\n" % (ftype, width, height))
    outfh.write(b"%d\n" % max_value)
    outfh.write(data)


def write_jbig2(outfh: BinaryIO, data: bytes, jbig2globals: bytes) -> None:
    """Write stream data to a JBIG2 file."""
    outfh.write(JBIG2_HEADER)
    # flags
    outfh.write(b"\x01")
    # number of pages
    outfh.write(b"\x00\x00\x00\x01")
    # write global segments
    outfh.write(jbig2globals)
    # write the rest of the data
    outfh.write(data)
    # and an eof segment
    outfh.write(
        b"\x00\x00\x00\x00"  # number (bogus!)
        b"\x33"  # flags: SEG_TYPE_END_OF_FILE
        b"\x00"  # retention_flags: empty
        b"\x00"  # page_assoc: 0
        b"\x00\x00\x00\x00"  # data_length: 0
    )


def write_cmyk_tiff(
    outfh: BinaryIO, data: bytes, bits: int, width: int, height: int
) -> None:
    """
    Writes a CMYK image to a TIFF file.
    """
    # 1. --- Rescale 1, 2, 4 bits to 8 bits ---
    # 1, 2, 4 BitsPerSample for CMYK not allowed as of TIFF Revision 6.0
    # Technically 16 BitsPerSample is also not allowed? but definitely exists
    # e.g. PIL can open them (albeit as 8 bit images)
    if bits in (1, 2, 4):
        scale_factor = 255 // ((1 << bits) - 1)  # 1, 3, 15 divides 255
        data = bytes(b * scale_factor for b in data)
        bits = 8

    # 2. --- TIFF Structure Constants & Calculations ---
    byte_order: Final = b"MM"  # Big-endian

    # --- Tag Codes ---
    TAG_IMAGE_WIDTH: Final = 256
    TAG_IMAGE_LENGTH: Final = 257
    TAG_BITS_PER_SAMPLE: Final = 258
    TAG_COMPRESSION: Final = 259
    TAG_PHOTOMETRIC_INTERP: Final = 262
    TAG_STRIP_OFFSETS: Final = 273
    TAG_SAMPLES_PER_PIXEL: Final = 277
    TAG_ROWS_PER_STRIP: Final = 278
    TAG_STRIP_BYTE_COUNTS: Final = 279
    TAG_X_RESOLUTION: Final = 282
    TAG_Y_RESOLUTION: Final = 283
    TAG_PLANAR_CONFIGURATION: Final = 284
    TAG_RESOLUTION_UNIT: Final = 296

    # --- Data Type Codes ---
    TYPE_SHORT: Final = 3
    TYPE_LONG: Final = 4
    TYPE_RATIONAL: Final = 5

    num_tags: Final = 13

    # --- Calculate Offsets ---
    # The file is laid out as: Header -> IFD -> Extra IFD Data -> Image Data
    header_size: Final = 8
    # 2 for tag count, 12 per tag, 4 for next IFD offset
    ifd_size: Final = 2 + (num_tags * 12) + 4

    # Offsets are relative to the start of the file
    offset_extra_data_start: Final = header_size + ifd_size
    offset_bits_per_sample: Final = offset_extra_data_start
    # 4 samples * 2 bytes/SHORT for the BitsPerSample array
    samples_per_pixel: Final = 4
    offset_x_resolution: Final = offset_bits_per_sample + (samples_per_pixel * 2)
    # 1 RATIONAL = 2 * 4 bytes
    offset_y_resolution: Final = offset_x_resolution + 8
    offset_image_data: Final = offset_y_resolution + 8

    # 3. --- Write TIFF Header ---
    # 8-byte header: Byte Order, TIFF Version (42), Offset to first IFD
    outfh.write(struct.pack(">2sHI", byte_order, 42, header_size))

    # 4. --- Write Image File Directory (IFD) ---
    # First, the number of tags in the directory
    outfh.write(struct.pack(">H", num_tags))

    # Write each 12-byte tag entry.
    # Format: Tag ID (2), Type (2), Count (4), Value/Offset (4)

    # Tag: ImageWidth (TYPE_SHORT)
    outfh.write(struct.pack(">HHII", TAG_IMAGE_WIDTH, TYPE_SHORT, 1, width << 16))

    # Tag: ImageLength (Height) (TYPE_SHORT)
    outfh.write(struct.pack(">HHII", TAG_IMAGE_LENGTH, TYPE_SHORT, 1, height << 16))

    # Tag: BitsPerSample (points to data written later)
    outfh.write(
        struct.pack(
            ">HHII",
            TAG_BITS_PER_SAMPLE,
            TYPE_SHORT,
            samples_per_pixel,
            offset_bits_per_sample,
        )
    )

    # Tag: Compression (1 = no compression)
    outfh.write(struct.pack(">HHII", TAG_COMPRESSION, TYPE_SHORT, 1, 1 << 16))

    # Tag: PhotometricInterpretation (5 = CMYK)
    outfh.write(struct.pack(">HHII", TAG_PHOTOMETRIC_INTERP, TYPE_SHORT, 1, 5 << 16))

    # Tag: StripOffsets (points to the single strip of image data)
    outfh.write(
        struct.pack(">HHII", TAG_STRIP_OFFSETS, TYPE_LONG, 1, offset_image_data)
    )

    # Tag: SamplesPerPixel
    outfh.write(
        struct.pack(
            ">HHII", TAG_SAMPLES_PER_PIXEL, TYPE_SHORT, 1, samples_per_pixel << 16
        )
    )

    # Tag: RowsPerStrip (the whole image is one strip)
    outfh.write(struct.pack(">HHII", TAG_ROWS_PER_STRIP, TYPE_LONG, 1, height))

    # Tag: StripByteCounts (total size of the image data)
    outfh.write(struct.pack(">HHII", TAG_STRIP_BYTE_COUNTS, TYPE_LONG, 1, len(data)))

    # Tag: XResolution (points to rational data written later)
    outfh.write(
        struct.pack(">HHII", TAG_X_RESOLUTION, TYPE_RATIONAL, 1, offset_x_resolution)
    )

    # Tag: YResolution (points to rational data written later)
    outfh.write(
        struct.pack(">HHII", TAG_Y_RESOLUTION, TYPE_RATIONAL, 1, offset_y_resolution)
    )

    # Tag: PlanarConfiguration (1 = chunky format CMYKCMYK...)
    outfh.write(struct.pack(">HHII", TAG_PLANAR_CONFIGURATION, TYPE_SHORT, 1, 1 << 16))

    # Tag: ResolutionUnit (2 = inches)
    outfh.write(struct.pack(">HHII", TAG_RESOLUTION_UNIT, TYPE_SHORT, 1, 2 << 16))

    # Write the offset to the next IFD (0 means this is the last one)
    outfh.write(struct.pack(">I", 0))

    # 5. --- Write Data Pointed to by IFD ---
    # BitsPerSample data: [bits, bits, bits, bits]
    outfh.write(struct.pack(">HHHH", bits, bits, bits, bits))

    # X and Y Resolution data (e.g., 72 DPI) as a RATIONAL (numerator, denominator)
    dpi = 72
    outfh.write(struct.pack(">II", dpi, 1))  # XResolution
    outfh.write(struct.pack(">II", dpi, 1))  # YResolution

    # 6. --- Write the Actual Pixel Data ---
    # The current file position should now match `offset_image_data`
    assert outfh.tell() == offset_image_data, (
        f"File position mismatch: at {outfh.tell()}, expected {offset_image_data}"
    )
    outfh.write(data)



"""
Test the PDF content stream interpreter.
"""

import logging
from pathlib import Path

import playa
from playa.document import LITERAL_TYPE1
from playa.interp import FontMapping, Type3Interpreter
from playa.parser import IndirectObjectParser

from .data import TESTDIR

THISDIR = Path(__file__).parent


def test_make_fontmap(caplog) -> None:
    """Test fontmap creation"""
    with playa.open(TESTDIR / "pdf_structure.pdf") as pdf:
        assert len(FontMapping([1, 2, 3], pdf)) == 0
        assert "not a dict" in caplog.text
        fonts = playa.resolve(pdf.pages[0].resources["Font"])
        fontmap = FontMapping(fonts, pdf)
        assert fontmap.keys() == {"F1", "F2"}
        fontmap = FontMapping(
            {"F1": {"Subtype": LITERAL_TYPE1, "FontDescriptor": 42}}, pdf
        )


def test_init_resources(caplog) -> None:
    with playa.open(THISDIR / "bad_resources.pdf") as pdf:
        for _ in pdf.pages[0]:
            pass
        assert "Missing" in caplog.text
        assert "not a dict" in caplog.text


def test_iter(caplog) -> None:
    caplog.set_level(logging.DEBUG)
    with playa.open(THISDIR / "bad_operators.pdf") as pdf:
        for _ in pdf.pages[0]:
            pass
    assert "Insufficient arguments" in caplog.text
    assert "Incorrect type" in caplog.text
    assert "Undefined xobject" in caplog.text
    assert "invalid xobject" in caplog.text
    assert "Unsupported XObject" in caplog.text
    assert "Undefined Font" in caplog.text
    assert "is not a name object" in caplog.text
    assert "Missing property list" in caplog.text


TYPE3_CHAR1 = b"""
99 0 obj
<</Length 0>>
stream
850 0 0 -200 1000 800 d1
% execute some color operators which do nothing
/Foo sh
0.5 scn
0.5 SCN
0.1 0.1 0.1 0.1 k
0.1 0.1 0.1 0.1 K
0.5 0.8 0.6 rg
0.5 0.8 0.6 RG
0.9 g
0.9 G
/DeviceRGB cs
/DeviceGray CS
/ReverseColorimetric ri
endstream
endobj
"""
TYPE3_CHAR2 = b"""
99 0 obj
<</Length 0>>
stream
800 0 d0
endstream
endobj
"""


def test_type3_interp() -> None:
    with playa.open(TESTDIR / "simple1.pdf") as pdf:
        _, obj = next(IndirectObjectParser(TYPE3_CHAR1))
        interp = Type3Interpreter(pdf.pages[0], [obj.obj])
        for _ in interp:
            pass
        assert interp.width == 850
        _, obj = next(IndirectObjectParser(TYPE3_CHAR2))
        interp = Type3Interpreter(pdf.pages[0], [obj.obj])
        for _ in interp:
            pass
        assert interp.width == 800
        # To get coverage
        assert interp.pop(0) == []



"""
Lazy interface to PDF document outline (PDF 1.7 sect 12.3.3).
"""

import logging
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Sequence,
    Type,
    Union,
)

from playa.exceptions import PDFSyntaxError
from playa.parser import PDFObject, PSLiteral
from playa.pdftypes import (
    LIT,
    ObjRef,
    Point,
    Rect,
    dict_value,
    num_value,
    point_value,
    rect_value,
    resolve1,
)
from playa.structure import Element
from playa.utils import apply_matrix_pt, decode_text, normalize_rect, transform_bbox
from playa.worker import (
    DocumentRef,
    PageRef,
    _deref_page,
    _deref_document,
    _ref_document,
)

if TYPE_CHECKING:
    from playa.document import Document
    from playa.page import Page

LOG = logging.getLogger(__name__)
DISPLAY_XYZ = LIT("XYZ")
DISPLAY_FIT = LIT("Fit")
DISPLAY_FITH = LIT("FitH")
DISPLAY_FITV = LIT("FitV")
DISPLAY_FITR = LIT("FitR")
DISPLAY_FITB = LIT("FitB")
DISPLAY_FITBH = LIT("FitBH")
DISPLAY_FITBV = LIT("FitBV")
ACTION_GOTO = LIT("GoTo")


@dataclass
class Destination:
    """PDF destinations (PDF 1.7 sect 12.3.2)"""

    _pageref: PageRef
    display: Union[PSLiteral, None]
    coords: List[PDFObject]
    ncoords: ClassVar[int] = 0

    @classmethod
    def from_dest(
        cls, dest: Union[PSLiteral, bytes, list], doc: "Document"
    ) -> "Destination":
        if isinstance(dest, (bytes, PSLiteral)):
            return doc.destinations[dest]
        elif isinstance(dest, list):
            return cls.from_list(doc, dest)
        else:
            raise TypeError("Unknown destination type: %r", dest)

    @classmethod
    def from_list(cls, doc: "Document", dest: Sequence) -> "Destination":
        pageobj, display, *coords = dest
        cls = DEST_CLASSES.get(display, cls)
        page = doc.pages[0]
        if isinstance(pageobj, int):
            # Not really sure if this is page number or page index...
            page = doc.pages[pageobj - 1]
        elif isinstance(pageobj, ObjRef):
            try:
                page = doc.pages.by_id(pageobj.objid)
            except KeyError:
                LOG.warning("Invalid page object in destination: %r", pageobj)
        else:
            LOG.warning("Unrecognized page in destination object: %r", pageobj)
        if not isinstance(display, PSLiteral):
            LOG.warning("Unknown display type: %r", display)
            display = None
        return cls(
            _pageref=page.pageref,
            display=display,
            coords=coords,
        )

    @property
    def page(self) -> "Page":
        """The page target of this destination."""
        return _deref_page(self._pageref)

    @property
    def doc(self) -> "Document":
        """The document containing this destination."""
        docref, _ = self._pageref
        return _deref_document(docref)

    @property
    def page_idx(self) -> int:
        return self.page.page_idx

    @property
    def top(self) -> Union[float, None]:
        """Top position for this destination, or None for unchanged."""
        return None

    @property
    def left(self) -> Union[float, None]:
        """Left position for this destination, or None for unchanged."""
        return None

    @property
    def pos(self) -> Point:
        """Position of this destination in device space.

        In the case where a coordinate is None (for "unchanged") it
        will default to the left or top edge of the page.
        """
        return apply_matrix_pt(self.page.ctm, (0, self.page.height))

    @property
    def bbox(self) -> Rect:
        """Rectangle in device space to zoom to for this destination.

        In the case where a coordinate is None (for "unchanged") it
        will default to the edge of the page.
        """
        left, top = self.pos
        right, bottom = apply_matrix_pt(self.page.ctm, (self.page.width, 0))
        return normalize_rect((left, top, right, bottom))

    @property
    def zoom(self) -> Union[float, None]:
        """Zoom factor of this destination, or None for unchanged."""
        return None


@dataclass
class DestinationXYZ(Destination):
    """Destination of type XYZ, with a (left, top) position and a zoom level."""

    ncoords: ClassVar[int] = 3

    @property
    def top(self) -> Union[float, None]:
        """Top position for this destination, or None for unchanged."""
        top = self.coords[1]
        if top is None:
            return None
        _, y = apply_matrix_pt(self.page.ctm, (0, num_value(top)))
        return y

    @property
    def left(self) -> Union[float, None]:
        """Left position for this destination, or None for unchanged."""
        left = self.coords[0]
        if left is None:
            return None
        x, _ = apply_matrix_pt(self.page.ctm, (num_value(left), 0))
        return x

    @property
    def pos(self) -> Point:
        """Position of this destination in device space.

        In the case where a coordinate is None (for "unchanged") it
        will default to the left or top edge of the page.
        """
        left, top = self.coords[0:2]
        # Give them some defaults in default user space
        if left is None:
            left = 0
        if top is None:
            top = self.page.height
        return apply_matrix_pt(self.page.ctm, point_value([left, top]))

    @property
    def zoom(self) -> Union[float, None]:
        """Zoom factor of this destination, or None for unchanged."""
        if not self.coords[2]:  # 0 and None are equivalent
            return None
        return num_value(self.coords[2])


@dataclass
class DestinationFitH(Destination):
    """Destination of type FitH or FitBH, with a top position."""

    ncoords: ClassVar[int] = 1

    @property
    def top(self) -> Union[float, None]:
        """Top position for this destination, or None for unchanged."""
        top = self.coords[0]
        if top is None:
            return None
        _, y = apply_matrix_pt(self.page.ctm, (0, num_value(top)))
        return y

    @property
    def left(self) -> Union[float, None]:
        """Left position for this destination (always None, unchanged)."""
        return None

    @property
    def pos(self) -> Point:
        """Position of this destination in device space.

        In the case where a coordinate is None (for "unchanged") it
        will default to the left or top edge of the page.
        """
        top = self.coords[0]
        left = 0
        if top is None:
            top = self.page.height
        return apply_matrix_pt(self.page.ctm, point_value([left, top]))


@dataclass
class DestinationFitV(Destination):
    """Destination of type FitV or FitBV, with a left position."""

    ncoords: ClassVar[int] = 1

    @property
    def top(self) -> Union[float, None]:
        """Top position for this destination (always None, unchanged)."""
        return None

    @property
    def left(self) -> Union[float, None]:
        """Left position for this destination, or None for unchanged."""
        left = self.coords[0]
        if left is None:
            return None
        x, _ = apply_matrix_pt(self.page.ctm, (num_value(left), self.page.height))
        return x

    @property
    def pos(self) -> Point:
        """Position of this destination in device space.

        In the case where a coordinate is None (for "unchanged") it
        will default to the left or top edge of the page.
        """
        left = self.coords[0]
        top = self.page.height
        if left is None:
            left = 0
        return apply_matrix_pt(self.page.ctm, point_value([left, top]))


@dataclass
class DestinationFitR(Destination):
    """Destination of type FitR, with a bounding box."""

    ncoords: ClassVar[int] = 4

    @property
    def top(self) -> Union[float, None]:
        """Top position for this destination, or None for unchanged."""
        top = self.coords[3]
        if top is None:
            return None
        _, y = self.pos
        return y

    @property
    def left(self) -> Union[float, None]:
        """Left position for this destination, or None for unchanged."""
        left = self.coords[0]
        if left is None:
            return None
        x, _ = self.pos
        return x

    @property
    def pos(self) -> Point:
        """Position of this destination in device space.

        In the case where a coordinate is None (for "unchanged") it
        will default to the left or top edge of the page.
        """
        left, _, _, top = self.coords
        # Give them some defaults in default user space
        if left is None:
            left = 0
        if top is None:
            top = self.page.height
        return apply_matrix_pt(self.page.ctm, point_value([left, top]))

    @property
    def bbox(self) -> Rect:
        """Rectangle in device space to zoom to for this destination.

        In the case where a coordinate is None (for "unchanged") it
        will default to the edge of the page.
        """
        left, bottom, right, top = self.coords
        if left is None:
            left = 0
        if bottom is None:
            bottom = 0
        if right is None:
            right = self.page.width
        if top is None:
            top = self.page.height
        return transform_bbox(self.page.ctm, rect_value([left, bottom, right, top]))


DEST_CLASSES: Dict[PSLiteral, Type[Destination]] = {
    DISPLAY_XYZ: DestinationXYZ,
    DISPLAY_FIT: Destination,
    DISPLAY_FITH: DestinationFitH,
    DISPLAY_FITV: DestinationFitV,
    DISPLAY_FITR: DestinationFitR,
    DISPLAY_FITB: Destination,
    DISPLAY_FITBH: DestinationFitH,
    DISPLAY_FITBV: DestinationFitV,
}


@dataclass
class Action:
    """PDF actions (PDF 1.7 sect 12.6)"""

    _docref: DocumentRef
    props: Dict[str, PDFObject]

    @classmethod
    def from_dict(cls, obj: PDFObject, doc: "Document") -> "Action":
        props = dict_value(obj)
        subtype = props.get("S")
        if subtype is None or not isinstance(subtype, PSLiteral):
            raise PDFSyntaxError("Invalid action subtype %r" % (subtype,))
        return cls(_docref=_ref_document(doc), props=props)

    @property
    def type(self) -> PSLiteral:
        assert isinstance(self.props["S"], PSLiteral)
        return self.props["S"]

    @property
    def doc(self) -> "Document":
        """Get associated document if it exists."""
        return _deref_document(self._docref)


class Tree(Iterable["Item"]):
    """PDF document outline (PDF 1.7 sect 12.3.3)"""

    _docref: DocumentRef
    props: Dict[str, PDFObject]

    def __init__(self, doc: "Document") -> None:
        self._docref = _ref_document(doc)
        self.props = dict_value(doc.catalog["Outlines"])

    def __iter__(self) -> Iterator["Item"]:
        if "First" in self.props and "Last" in self.props:
            ref = self.props["First"]
            while ref is not None:
                if not isinstance(ref, ObjRef):
                    LOG.warning("Not an indirect object reference: %r", ref)
                    break
                out = Item(_docref=self._docref, props=dict_value(ref))
                ref = out.props.get("Next")
                yield out
                if ref is self.props["Last"]:
                    break

    @property
    def doc(self) -> "Document":
        """Get associated document if it exists."""
        return _deref_document(self._docref)


@dataclass
class Item(Iterable["Item"]):
    """PDF document outline item (PDF 1.7 sect 12.3.3)"""

    _docref: DocumentRef
    props: Dict[str, PDFObject]

    def __iter__(self) -> Iterator["Item"]:
        if "First" in self.props and "Last" in self.props:
            ref = self.props["First"]
            while ref is not None:
                if not isinstance(ref, ObjRef):
                    LOG.warning("Not an indirect object reference: %r", ref)
                    break
                out = Item(_docref=self._docref, props=dict_value(ref))
                ref = out.props.get("Next")
                yield out
                if ref is self.props["Last"]:
                    break

    @property
    def doc(self) -> "Document":
        """Get associated document if it exists."""
        return _deref_document(self._docref)

    @property
    def title(self) -> str:
        raw = resolve1(self.props.get("Title"))
        if raw is None:
            raise PDFSyntaxError(f"Outline item has no title: {self.props!r}")
        if not isinstance(raw, bytes):
            raise PDFSyntaxError(f"Outline item title is not a string: {raw!r}")
        return decode_text(raw)

    @property
    def destination(self) -> Union[Destination, None]:
        """Destination for this outline item.

        Note: Special case of `GoTo` actions.
            Since internal `GoTo` actions (PDF 1.7 sect 12.6.4.2) in
            outlines and links are entirely equivalent to
            destinations, if one exists, it will be returned here as
            well.

        Returns:
            destination, if one exists.
        """
        dest = resolve1(self.props.get("Dest"))
        if dest is not None:
            try:
                if isinstance(dest, (PSLiteral, bytes, list)):
                    return Destination.from_dest(dest, self.doc)
            except KeyError:
                LOG.warning("Unknown named destination: %r", dest)
        # Fall through to try an Action instead
        action = self.action
        if action is None or action.type is not ACTION_GOTO:
            return None
        try:
            dest = action.props["D"]
            if isinstance(dest, (PSLiteral, bytes, list)):
                return Destination.from_dest(dest, self.doc)
        except KeyError:
            LOG.warning("Unknown named destination: %r", dest)
        return None

    @property
    def action(self) -> Union[Action, None]:
        try:
            return Action(self._docref, dict_value(self.props["A"]))
        except (KeyError, TypeError):
            return None

    @property
    def element(self) -> Union[Element, None]:
        """The structure element associated with this outline item, if
        any.

        Returns:
            structure element, if one exists.
        """
        try:
            return Element.from_dict(self.doc, dict_value(self.props["SE"]))
        except (KeyError, TypeError):
            return None

    @property
    def parent(self) -> Union["Item", "Tree"]:
        ref = self.props["Parent"]
        if not isinstance(ref, ObjRef):
            raise PDFSyntaxError(f"Parent is not indirect object reference: {ref!r}")
        props = dict_value(ref)
        if "Parent" not in props:
            tree = self.doc.outline
            if tree is None:
                raise ValueError("Document apparently has no outline?!?")
        return Item(self._docref, props)



"""
Test document outline.
"""

from typing import List

import pytest

import playa
from playa.exceptions import PDFEncryptionError, PDFSyntaxError
from playa.outline import Tree as OutlineTree, Item as OutlineItem

from .data import ALLPDFS, PASSWORDS, TESTDIR, XFAILS


def expand_titles(outline: OutlineTree) -> List:
    def expand_one(child: OutlineItem):
        out = [child.title]
        for c in child:
            out.append(expand_one(c))
        return out

    out = []
    for child in outline:
        out.extend(expand_one(child))
    return out


def test_outline():
    """Test basic outline functionality."""
    with playa.open(TESTDIR / "pdf_structure.pdf") as pdf:
        titles = expand_titles(pdf.outline)
        assert titles == ["Titre 1", ["Titre 2", ["Tableau"]]]


def expand(outline: OutlineTree) -> List:
    def expand_one(child: OutlineItem, level=1):
        out = [child.title, child.destination, child.element]
        # Limit depth to avoid taking all memory
        if level == 3:
            return out
        for c in child:
            out.append(expand_one(c, level + 1))
        return out

    out = []
    for idx, child in enumerate(outline):
        # Limit number to avoid taking all memory
        if idx == 10:
            break
        out.extend(expand_one(child))
    return out


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_outlines(path) -> None:
    """Verify that we can read outlines when they exist."""
    if path.name in XFAILS:
        pytest.xfail("Intentionally corrupt file: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            with playa.open(path, password=password) as pdf:
                outline = pdf.outline
                if outline is not None:
                    print(expand(outline))
        except PDFEncryptionError:
            pytest.skip("password incorrect or cryptography package not installed")
        except PDFSyntaxError as e:
            # Make sure we report the error as an invalid outline
            assert "outline" in str(e).lower()


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_destinations(path) -> None:
    """Verify that we can read destinations when they exist."""
    if path.name in XFAILS:
        pytest.xfail("Intentionally corrupt file: %s" % path.name)
    if path.name == "issue2098.pdf":
        pytest.skip("Skipping excessively borken PDF with bad destinations")
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            with playa.open(path, password=password) as pdf:
                for idx, k in enumerate(pdf.destinations):
                    dest = pdf.destinations[k]
                    assert dest.pos
                    # FIXME: Currently getting destinations is quite
                    # slow, so only do a few of them.
                    if idx == 10:
                        break
        except PDFEncryptionError:
            pytest.skip("password incorrect or cryptography package not installed")



"""
Test the varieties of cross-reference table as implemented in playa.xref
"""

from dataclasses import dataclass
from pathlib import Path
from typing import cast

import playa
import pytest
from playa.document import Document
from playa.exceptions import PDFSyntaxError
from playa.parser import ObjectParser
from playa.pdftypes import LIT
from playa.xref import XRefFallback, XRefStream, XRefTable

from .data import CONTRIB, TESTDIR

THISDIR = Path(__file__).parent


def test_read_xref():
    """Verify we can read the xref table if there is junk before the header."""
    with playa.open(TESTDIR / "junk_before_header.pdf") as pdf:
        # Not a fallback, we got the right one
        assert isinstance(pdf.xrefs[0], XRefTable)

        # Verify that the positions are the file positions
        assert pdf.xrefs[0][1].pos == 9 + pdf._offset
        assert pdf.xrefs[0][6].pos == 954 + pdf._offset

        # Verify that we can get data
        assert pdf[2] == {"Type": LIT("Outlines"), "Count": 0}

        # Verify that free objects are free
        with pytest.raises(KeyError):
            pdf[42]


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_root_damage() -> None:
    """Fail gracefully if the document root is damaged (issue #154)"""
    with playa.open(CONTRIB / "issue-154.pdf") as doc:
        assert isinstance(doc[827], playa.ContentStream)


def test_multi_xrefs(caplog) -> None:
    """Verify that we correctly read multi-segment xref tables."""
    with playa.open(TESTDIR / "multi-xrefs.pdf"):
        assert not caplog.records


# OMGFU GIT AND YOUR CRLF TRANSLATION
GOOD_XREF1 = (
    b"0 3\n"
    b"0000000000 65535 f \n"
    b"0000000010 00000 n \n"
    b"0000000020 00000 n \n"
    b"5 2\n"
    b"0000000030 00000 n \n"
    b"0000000040 00000 n \n"
    b"trailer\n"
    b"<</Size 7 /Root 1 0 R>>\n"
    b"startxref\n"
    b"0\n"
    b"%%EOF\n"
)


@dataclass
class MockDoc:
    buffer: bytes

    def decipher(self, _objid, _genno, data, *args, **kwargs):
        return data


def mock_doc(buffer: bytes) -> Document:
    return cast(Document, MockDoc(buffer=buffer))


def test_xref_tables() -> None:
    """Verify that we can read valid xref tables."""
    x = XRefTable(mock_doc(GOOD_XREF1))
    assert repr(x)
    assert [1, 2, 5, 6] == list(x)
    crlf = GOOD_XREF1.replace(b" \n", b"\r\n")
    x = XRefTable(mock_doc(crlf))
    assert [1, 2, 5, 6] == list(x)
    cr = GOOD_XREF1.replace(b" \n", b" \r")
    x = XRefTable(mock_doc(cr))
    assert [1, 2, 5, 6] == list(x)


# EOF before trailer (no trailer = fallback)
BAD_XREF1 = (
    b"0 3\n"
    b"0000000000 65535 f \n"
    b"0000000010 00000 n \n"
    b"0000000020 00000 n \n"
    b"5 2\n"
    b"0000000030 00000 n \n"
    b"0000000040 00000 n \n"
)


# EOF in table
BAD_XREF2 = b"0 5\n0000000000 65535 f \n0000000010 00000 n \n0000000020 00000 n \n"


# Junk in table
BAD_XREF3 = (
    b"0 5\n0000000000 65535 f FOOBIE\n0000000010 00000 n BLETCH\n0000000020 00000 n \n"
)


# Missing "trailer"
BAD_XREF4 = (
    b"0 3\n"
    b"0000000000 65535 f \n"
    b"0000000010 00000 n \n"
    b"0000000020 00000 n \n"
    b"5 2\n"
    b"0000000030 00000 n \n"
    b"0000000040 00000 n \n"
    b"not_a_trailer\n"
    b"<</Size 7 /Root 1 0 R>>\n"
    b"startxref\n"
    b"0\n"
    b"%%EOF\n"
)


def test_bad_xref_tables() -> None:
    """Verify that we fail on fatally flawed xref tables."""
    with pytest.raises(StopIteration):
        XRefTable(mock_doc(BAD_XREF1))
    with pytest.raises(PDFSyntaxError):
        XRefTable(mock_doc(BAD_XREF2))
    with pytest.raises(PDFSyntaxError):
        XRefTable(mock_doc(BAD_XREF3))
    with pytest.raises(PDFSyntaxError):
        XRefTable(mock_doc(BAD_XREF4))
    with pytest.raises(PDFSyntaxError):
        x = XRefTable(mock_doc(GOOD_XREF1))
        x._load_trailer(ObjectParser(b"not_a_trailer"))


# F***ed up whitespace but still readable
UGLY_XREF1 = (
    b"0 3 0000000000 65535 f\n"
    b"\n"
    b"0000000010 00000 n\n"
    b"\n"
    b"0000000020 00000 n\n"
    b"\n"
    b"5 2 0000000030 00000 n\n"
    b"0000000040 00000 n\n"
    b"trailer\n"
    b"<</Size 7 /Root 1 0 R>>\n"
    b"startxref\n"
    b"0\n"
    b"%%EOF\n"
)
# FIXME: Don't yet handle the case of 20-byte records with no newlines

# Object count is greater than number of records
UGLY_XREF2 = (
    b"0 5\n"
    b"0000000000 65535 f \n"
    b"0000000010 00000 n \n"
    b"0000000020 00000 n \n"
    b"trailer\n"
    b"<</Size 3 /Root 1 0 R>>\n"
    b"startxref\n"
    b"0\n"
    b"%%EOF\n"
)
# FIXME: Don't yet handle the case of too small nobjs


XREF_STREAM1 = b"""1 0 obj
<</Type/XRef/Size 5/W[1 3 2]/Filter[/ASCIIHexDecode]/Length 75>>
stream
00 ffffff 0000
01 000010 0000
01 000020 0000
01 000030 0000
01 000040 0000
endstream
endobj
"""


def test_xref_streams() -> None:
    """Verify that we can read xref streams."""
    s = XRefStream(mock_doc(XREF_STREAM1))
    assert repr(s)
    assert list(s) == [1, 2, 3, 4]
    assert s[2].pos == 32
    with pytest.raises(KeyError):
        s[0]


BAD_XREF_STREAM1 = b"""1 0 obj
<</Type/Other/Size 5/W[1 3 2]/Filter[/ASCIIHexDecode]/Length 75>>
stream
00 ffffff 0000
01 000010 0000
01 000020 0000
01 000030 0000
01 000040 0000
endstream
endobj
"""


BAD_XREF_STREAM2 = b"""1 0 obj
<</Type/XRef/Size 5/Index[0 5 3]/W[1 3 2]/Filter[/ASCIIHexDecode]/Length 75>>
stream
00 ffffff 0000
01 000010 0000
01 000020 0000
01 000030 0000
01 000040 0000
endstream
endobj
"""


def test_bad_xref_streams() -> None:
    """Verify that we reject bad xref streams."""
    with pytest.raises(ValueError):
        XRefStream(mock_doc(BAD_XREF_STREAM1))
    with pytest.raises(PDFSyntaxError):
        XRefStream(mock_doc(BAD_XREF_STREAM2))


def test_xref_fallback() -> None:
    """Reconstruct xref table from a test document."""

    data = (THISDIR / "fallback-xref.pdf").read_bytes()
    f = XRefFallback(mock_doc(data))
    assert repr(f)
    pos2 = f[2]
    assert pos2.genno == 1
    assert data[pos2.pos :].startswith(b"2 1 obj")
    assert list(f) == [1, 2, 3, 4, 7, 6, 5]
    pos7 = f[7]
    assert pos7.streamid == 3
    assert f.trailer == {"Root": 1, "Size": 7}
    f = XRefFallback(mock_doc(data))
